###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.JobStepPlotter.

JobStepPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from DIRAC import S_OK
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.JobStep import JobStep


class JobStepPlotter(BaseReporter):
    """JobStepPlotter as an extension of BaseReporter."""

    _typeName = "JobStep"
    _typeKeyFields = [dF[0] for dF in JobStep().definitionKeyFields]

    # .............................................................................
    # CPU Efficiency

    _reportCPUEfficiencyName = "CPU efficiency"

    def _reportCPUEfficiency(self, reportRequest):
        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s), SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "CPUTime", "ExecTime"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {
                "checkNone": True,
                "convertToGranularity": "sum",
                "calculateProportionalGauges": False,
                "consolidationFunction": self._efficiencyConsolidation,
            },
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        if len(dataDict) > 1:
            # Get the total for the plot
            selectFields = ("'Total', %s, %s, SUM(%s),SUM(%s)", ["startTime", "bucketLength", "CPUTime", "ExecTime"])

            retVal = self._getTimedData(
                reportRequest["startTime"],
                reportRequest["endTime"],
                selectFields,
                reportRequest["condDict"],
                reportRequest["groupingFields"],
                {
                    "scheckNone": True,
                    "convertToGranularity": "sum",
                    "calculateProportionalGauges": False,
                    "consolidationFunction": self._efficiencyConsolidation,
                },
            )
            if not retVal["OK"]:
                return retVal
            totalDict = retVal["Value"][0]
            self.stripDataField(totalDict, 0)
            for key in totalDict:
                dataDict[key] = totalDict[key]

        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotCPUEfficiency(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": f"CPU efficiency by {reportRequest['grouping']}",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
        }

        return self._generateQualityPlot(filename, plotInfo["data"], metadata)

    # .............................................................................
    # CPU Usage

    _reportCPUUsageName = "CPU time"

    def _reportCPUUsage(self, reportRequest):
        field = "CPUTime"
        unit = "time"

        return self.__reportNormPlot(reportRequest, field, unit)

    def _plotCPUUsage(self, reportRequest, plotInfo, filename):
        title = "CPU usage"

        return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Pie CPU Time

    _reportPieCPUTimeName = "Pie plot of CPU time"

    def _reportPieCPUTime(self, reportRequest):
        return self.__reportPiePlot(reportRequest, "CPUTime")

    def _plotPieCPUTime(self, reportRequest, plotInfo, filename):
        title = "CPU time"
        label = "CPUTime"

        return self.__plotPie(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Cumulative CPU Time

    _reportCumulativeCPUTimeName = "Cumulative CPU time"

    def _reportCumulativeCPUTime(self, reportRequest):
        field = "CPUTime"
        unit = "time"

        return self.__reportCumulativePlot(reportRequest, field, unit)

    def _plotCumulativeCPUTime(self, reportRequest, plotInfo, filename):
        title = "Cumulative CPU time"

        return self.__plotCumulative(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Norm CPU Time

    _reportNormCPUTimeName = "NormCPU time"

    def _reportNormCPUTime(self, reportRequest):
        field = "NormCPUTime"
        unit = "time"

        return self.__reportNormPlot(reportRequest, field, unit)

    def _plotNormCPUTime(self, reportRequest, plotInfo, filename):
        title = "Normalized CPU"

        return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Cumulative Norm CPU Time

    _reportCumulativeNormCPUTimeName = "Cumulative normalized CPU time"

    def _reportCumulativeNormCPUTime(self, reportRequest):
        field = "NormCPUTime"
        unit = "time"

        return self.__reportCumulativePlot(reportRequest, field, unit)

    def _plotCumulativeNormCPUTime(self, reportRequest, plotInfo, filename):
        title = "Cumulative Normalized CPU time"

        return self.__plotCumulative(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Pie Norm CPU Time

    _reportPieNormCPUTimeName = "Pie plot of NormCPU time"

    def _reportPieNormCPUTime(self, reportRequest):
        return self.__reportPiePlot(reportRequest, "NormCPUTime")

    def _plotPieNormCPUTime(self, reportRequest, plotInfo, filename):
        title = "Average Normalized CPU time"
        label = "NormCPUTime"

        return self.__plotPie(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Input Data

    _reportInputDataName = "Input Data"

    def _reportInputData(self, reportRequest):
        field = "InputData"
        unit = "files"

        return self.__reportNormPlot(reportRequest, field, unit)

    def _plotInputData(self, reportRequest, plotInfo, filename):
        title = "Input data"

        return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Cumulative Input Data

    _reportCumulativeInputDataName = "Cumulative Input Data"

    def _reportCumulativeInputData(self, reportRequest):
        field = "InputData"
        unit = "files"

        return self.__reportCumulativePlot(reportRequest, field, unit)

    def _plotCumulativeInputData(self, reportRequest, plotInfo, filename):
        title = "Cumulative Input Data"

        return self.__plotCumulative(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Pie Input Data

    _reportPieInputDataName = "Pie plot of Input Data"

    def _reportPieInputData(self, reportRequest):
        return self.__reportPiePlot(reportRequest, "InputData")

    def _plotPieInputData(self, reportRequest, plotInfo, filename):
        title = "Pie plot of Input data"
        label = "InputData"

        return self.__plotPie(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Output Data

    _reportOutputDataName = "Output Data"

    def _reportOutputData(self, reportRequest):
        field = "OutputData"
        units = "files"

        return self.__reportNormPlot(reportRequest, field, units)

    def _plotOutputData(self, reportRequest, plotInfo, filename):
        title = "Output data"

        return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Cumulative Output Data

    _reportCumulativeOutputDataName = "Cumulative OutputData"

    def _reportCumulativeOutputData(self, reportRequest):
        field = "OutputData"
        unit = "files"

        return self.__reportCumulativePlot(reportRequest, field, unit)

    def _plotCumulativeOutputData(self, reportRequest, plotInfo, filename):
        title = "Cumulative Output Data"

        return self.__plotCumulative(reportRequest, plotInfo, filename, title)

    # .............................................................................
    # Pie Output Data

    _reportPieOutputDataName = "Pie plot of Output Data"

    def _reportPieOutputData(self, reportRequest):
        return self.__reportPiePlot(reportRequest, "OutputData")

    def _plotPieOutputData(self, reportRequest, plotInfo, filename):
        title = "Pie plot of  Output data"
        label = "OutputData"

        return self.__plotPie(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Input Events

    _reportInputEventsName = "Input Events"

    def _reportInputEvents(self, reportRequest):
        return self.__reportNumberOfField(reportRequest, "InputEvents")

    def _plotInputEvents(self, reportRequest, plotInfo, filename):
        title = "InputEvents"
        label = "Input Events"

        return self.__plotNumberOfField(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Cumulative Input Events

    _reportCumulativeInputEventsName = "Cumulative Input Events"

    def _reportCumulativeInputEvents(self, reportRequest):
        return self.__reportCumulativeNumberOfField(reportRequest, "InputEvents")

    def _plotCumulativeInputEvents(self, reportRequest, plotInfo, filename):
        title = "Cumulative Input Events "
        label = "Input Events"

        return self.__plotCumulativeNumberOfField(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Pie Input Events

    _reportPieInputEventsName = "Pie plot of Input Events"

    def _reportPieInputEvents(self, reportRequest):
        return self.__reportPiePlot(reportRequest, "InputEvents")

    def _plotPieInputEvents(self, reportRequest, plotInfo, filename):
        title = "Pie plot of Input Events"
        label = "InputEvents"

        return self.__plotPie(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Output Events

    _reportOutputEventsName = "Output Events"

    def _reportOutputEvents(self, reportRequest):
        return self.__reportNumberOfField(reportRequest, "OutputEvents")

    def _plotOutputEvents(self, reportRequest, plotInfo, filename):
        title = "OutputEvents"
        label = "Output Events"

        return self.__plotNumberOfField(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Cumulative Output Events

    _reportCumulativeOutputEventsName = "Cumulative Output Events"

    def _reportCumulativeOutputEvents(self, reportRequest):
        return self.__reportCumulativeNumberOfField(reportRequest, "OutputEvents")

    def _plotCumulativeOutputEvents(self, reportRequest, plotInfo, filename):
        title = "Cumulative Output Events "
        label = "Output Events"

        return self.__plotCumulativeNumberOfField(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Pie Output Events

    _reportPieOutputEventsName = "Pie plot of Output Events"

    def _reportPieOutputEvents(self, reportRequest):
        return self.__reportPiePlot(reportRequest, "OutputEvents")

    def _plotPieOutputEvents(self, reportRequest, plotInfo, filename):
        title = "Pie plot of Output Events"
        label = "OutputEvents"

        return self.__plotPie(reportRequest, plotInfo, filename, title, label)

    # .............................................................................
    # Input Events Per Output Events

    _reportInputEventsPerOutputEventsName = "Input/Output Events"

    def _reportInputEventsPerOutputEvents(self, reportRequest):
        return self.__report2D(reportRequest, "InputEvents", "OutputEvents")

    def _plotInputEventsPerOutputEvents(self, reportRequest, plotInfo, filename):
        return self.__plot2D(reportRequest, plotInfo, filename, "Input/Output Events")

    # .............................................................................
    # CPU Time Per Output Events

    _reportCPUTimePerOutputEventsName = "CPUTime/Output Events"

    def _reportCPUTimePerOutputEvents(self, reportRequest):
        return self.__report2D(reportRequest, "CPUTime", "OutputEvents")

    def _plotCPUTimePerOutputEvents(self, reportRequest, plotInfo, filename):
        return self.__plot2D(reportRequest, plotInfo, filename, "CPUTime/Output Events")

    # .............................................................................
    # CPU Time Per Input Events

    _reportCPUTimePerInputEventsName = "CPUTime/Input Events"

    def _reportCPUTimePerInputEvents(self, reportRequest):
        return self.__report2D(reportRequest, "CPUTime", "InputEvents")

    def _plotCPUTimePerInputEvents(self, reportRequest, plotInfo, filename):
        return self.__plot2D(reportRequest, plotInfo, filename, "CPUTime/Input Events")

    # .............................................................................
    # HELPER methods

    def __reportNormPlot(self, reportRequest, field, unit):
        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", field],
        )

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]

        retVal = self._getTimedData(
            startTime, endTime, selectFields, reportRequest["condDict"], reportRequest["groupingFields"], {}
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        # 2nd element ( maxValue ) not used
        dataDict, __ = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)

        accumMaxVal = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableRateUnit(dataDict, accumMaxVal, unit)

        # 3rd element ( maxValue ) not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def __plotNormPlot(self, reportRequest, plotInfo, filename, title):
        metadata = {
            "title": f"{title} by {reportRequest['grouping']}",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }

        return self._generateStackedLinePlot(filename, plotInfo["graphDataDict"], metadata)

    def __reportPiePlot(self, reportRequest, field):
        # selectFields = (self._getSelectStringForGrouping(reportRequest[ 'groupingFields' ]) + ", SUM(%s/%s)",
        #                 reportRequest[ 'groupingFields' ][1] + [ field, 'entriesInBucket'
        #                               ]
        #               )

        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (selectField + ", SUM(%s)", reportRequest["groupingFields"][1] + [field])

        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        # bins = self._getBins(self._typeName, reportRequest[ 'startTime' ], reportRequest[ 'endTime' ])
        # numBins = len(bins)
        # for key in dataDict:
        #   dataDict[ key ] = float(dataDict[ key ] / numBins)
        return S_OK({"data": dataDict})

    def __plotPie(self, reportRequest, plotInfo, filename, title, label):
        metadata = {
            "title": f"{title} by {reportRequest['grouping']}",
            "ylabel": label,
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)

    def __reportCumulativePlot(self, reportRequest, field, unit):
        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", field],
        )

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]

        retVal = self._getTimedData(
            startTime, endTime, selectFields, reportRequest["condDict"], reportRequest["groupingFields"], {}
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)
        dataDict = self._accumulate(granularity, startTime, endTime, dataDict)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, unit)

        # 3rd element ( maxValue ) unused
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def __plotCumulative(self, reportRequest, plotInfo, filename, title):
        metadata = {
            "title": f"{title} by {reportRequest['grouping']}",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
            "sort_labels": "last_value",
        }

        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    def __reportNumberOfField(self, reportRequest, field):
        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", field],
        )

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]

        retVal = self._getTimedData(
            startTime, endTime, selectFields, reportRequest["condDict"], reportRequest["groupingFields"], {}
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        # 2nd element ( maxValue ) unused
        dataDict, __ = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableRateUnit(dataDict, accumMaxValue, "files")

        # 3rd element ( maxValue ) unused
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def __plotNumberOfField(self, reportRequest, plotInfo, filename, title, label):
        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["data"]

        metadata = {
            "title": f"{title}  by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "skipEdgeColor": True,
            "ylabel": label,
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    def __reportCumulativeNumberOfField(self, reportRequest, field):
        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", field],
        )

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]

        retVal = self._getTimedData(
            startTime,
            endTime,
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)
        dataDict = self._accumulate(granularity, startTime, endTime, dataDict)

        return S_OK({"data": dataDict, "granularity": granularity})

    def __plotCumulativeNumberOfField(self, reportRequest, plotInfo, filename, title, label):
        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["data"]

        metadata = {
            "title": f"{title}  by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "skipEdgeColor": True,
            "ylabel": label,
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # FIXME: smells like unused
    def __reportAverageNumberOfField(self, reportRequest, field):
        # selectFields = (self._getSelectStringForGrouping(reportRequest[ 'groupingFields' ]) + ",SUM(%s/%s)",
        #                 reportRequest[ 'groupingFields' ][1] + [Field, 'entriesInBucket'
        #                               ]
        #               )

        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (selectField + ",SUM(%s)", reportRequest["groupingFields"][1] + [field])

        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal

        dataDict = retVal["Value"]
        # bins = self._getBins(self._typeName, reportRequest[ 'startTime' ], reportRequest[ 'endTime' ])
        # numBins = len(bins)
        # for key in dataDict:
        #  dataDict[ key ] = float(dataDict[ key ] / numBins)
        return S_OK({"data": dataDict})

    # FIXME: smells like unused
    def __plotAverageNumberOfField(self, reportRequest, plotInfo, filename, title, label):
        metadata = {
            "title": f"{title} by {reportRequest['grouping']}",
            "ylabel": label,
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)

    def __report2D(self, reportRequest, field1, field2):
        selectField = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectField + ", %s, %s, SUM(%s), SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", field1, field2],
        )

        startTime = reportRequest["startTime"]
        endTime = reportRequest["endTime"]

        retVal = self._getTimedData(
            startTime,
            endTime,
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"checkNone": True, "convertToGranularity": "sum", "calculateProportionalGauges": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)
        return S_OK({"data": dataDict, "granularity": granularity})

    def __plot2D(self, reportRequest, plotInfo, filename, label):
        metadata = {
            "title": f"Jobs by {reportRequest['grouping']}",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": label,
        }

        return self._generateTimedStackedBarPlot(filename, plotInfo["data"], metadata)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module to upload specified job output files according to the parameters
defined in the user workflow."""
import os
import random
import re

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities.File import getGlobbedFiles
from DIRAC.DataManagementSystem.Client.FailoverTransfer import FailoverTransfer
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

from LHCbDIRAC.Core.Utilities.ProductionData import constructUserLFNs
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.ResolveSE import getDestinationSEList


class UserJobFinalization(ModuleBase):
    """Finalization of user jobs."""

    #############################################################################
    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""

        self.log = gLogger.getSubLogger("UserJobFinalization")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.enable = True
        self.defaultOutputSE = resolveSEGroup(gConfig.getValue("/Resources/StorageElementGroups/Tier1-USER", []))
        self.failoverSEs = resolveSEGroup(gConfig.getValue("/Resources/StorageElementGroups/Tier1-Failover", []))
        # List all parameters here
        self.request = None
        # Always allow any files specified by users
        self.outputDataFileMask = ""
        self.userOutputData = []
        self.userOutputSE = ""
        self.userOutputPath = ""
        self.failoverTransfer = None
        self.replicateUserOutputData = False
        self.userPrependString = ""

    #############################################################################
    def _resolveInputVariables(self):
        """By convention the module parameters are resolved here."""
        super()._resolveInputVariables()

        # Use LHCb utility for local running via dirac-jobexec
        if "UserOutputData" in self.workflow_commons:
            userOutputData = self.workflow_commons["UserOutputData"]
            if not isinstance(userOutputData, list):
                userOutputData = [i.strip() for i in userOutputData.split(";")]
            self.userOutputData = userOutputData

        if "UserOutputSE" in self.workflow_commons:
            specifiedSE = self.workflow_commons["UserOutputSE"]
            if not isinstance(specifiedSE, list):
                self.userOutputSE = [i.strip() for i in specifiedSE.split(";")]
        else:
            self.log.verbose(f"No UserOutputSE specified, using default value: {', '.join(self.defaultOutputSE)}")
            self.userOutputSE = []

        if "UserOutputPath" in self.workflow_commons:
            self.userOutputPath = self.workflow_commons["UserOutputPath"]

        if "ReplicateUserOutputData" in self.workflow_commons and self.workflow_commons["ReplicateUserOutputData"]:
            self.replicateUserOutputData = True

        if "UserOutputLFNPrepend" in self.workflow_commons:
            self.userPrependString = self.workflow_commons["UserOutputLFNPrepend"]

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
        orderedSEs=None,
    ):
        """Main execution function."""

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            self._resolveInputVariables()

            # Earlier modules may have populated the report objects
            self.request.RequestName = "job_%d_request.xml" % self.jobID
            self.request.JobID = self.jobID
            self.request.SourceComponent = "Job_%d" % self.jobID

            if not self._checkWFAndStepStatus():
                return S_OK()

            if not self.userOutputData:
                self.log.info("No user output data is specified for this job, nothing to do")
                return S_OK("No output data to upload")

            self.log.info(f"User specified output file list is: {', '.join(self.userOutputData)}")

            globList = []
            for i in self.userOutputData:
                if re.search(r"\*", i):
                    globList.append(i)

            # Check whether list of userOutputData is a globbable pattern
            if globList:
                for i in globList:
                    self.userOutputData.remove(i)

                globbedOutputList = list(set(getGlobbedFiles(globList)))
                if globbedOutputList:
                    self.log.info(
                        "Found a pattern in the output data file list, \
          extra files to upload are: %s"
                        % (", ".join(globbedOutputList))
                    )
                    self.userOutputData += globbedOutputList
                else:
                    self.log.info(
                        f"No files were found on the local disk for the following patterns: {', '.join(globList)}"
                    )

            self.log.info(f"Final list of files to upload are: {', '.join(self.userOutputData)}")

            # Determine the final list of possible output files for the workflow
            # and all the parameters needed to upload them.
            outputList = []
            for i in self.userOutputData:
                outputList.append({"outputDataType": (".".split(i)[-1]).upper(), "outputDataName": os.path.basename(i)})

            userOutputLFNs = []
            if self.userOutputData:
                self.log.info(f"Constructing user output LFN(s) for {', '.join(self.userOutputData)}")

                userOutputLFNs = constructUserLFNs(
                    self.jobID,
                    self._getCurrentOwner(),
                    self.userOutputData,
                    self.userOutputPath,
                    self.userPrependString,
                )

            self.log.verbose(f"Calling getCandidateFiles( {outputList}, {userOutputLFNs}, {self.outputDataFileMask})")
            try:
                fileDict = self.getCandidateFiles(outputList, userOutputLFNs, self.outputDataFileMask)
            except OSError as e:
                self.setApplicationStatus(e)
                return S_OK()

            try:
                fileMetadata = self.getFileMetadata(fileDict)
            except RuntimeError as e:
                self.setApplicationStatus(e)
                return S_OK()

            if not fileMetadata:
                self.log.info("No output data files were determined to be uploaded for this workflow")
                self.setApplicationStatus("No Output Data Files To Upload")
                return S_OK()

            if not orderedSEs:
                orderedSEs = self._getOrderedSEsList()

            self.log.info(f"Ordered list of output SEs is: {', '.join(orderedSEs)}")
            final = {}
            for fileName, metadata in fileMetadata.items():
                final[fileName] = metadata
                final[fileName]["resolvedSE"] = orderedSEs

            # At this point can exit and see exactly what the module will upload
            if not self._enableModule():
                self.log.info(f"Module disabled would have attempted to upload the files {', '.join(final)}")
                for fileName, metadata in final.items():
                    self.log.info(f"--------{fileName}--------")
                    for n, v in metadata.items():
                        self.log.info(f"{n} = {v}")

                return S_OK("Module is disabled by control flag")

            # Disable the watchdog check in case the file uploading takes a long time
            self._disableWatchdogCPUCheck()

            # Instantiate the failover transfer client with the global request object
            if not self.failoverTransfer:
                self.failoverTransfer = FailoverTransfer(self.request)

            # One by one upload the files with failover if necessary
            replication = {}
            failover = {}
            uploaded = []
            for fileName, metadata in final.items():
                self.log.info(
                    f"Attempting to store {fileName} to the following SE(s): {', '.join(metadata['resolvedSE'])}"
                )
                fileMetaDict = {
                    "Size": metadata["filedict"]["Size"],
                    "LFN": metadata["filedict"]["LFN"],
                    "GUID": metadata["filedict"]["GUID"],
                    "Checksum": metadata["filedict"]["Checksum"],
                    "ChecksumType": metadata["filedict"]["ChecksumType"],
                }
                result = self.failoverTransfer.transferAndRegisterFile(
                    fileName=fileName,
                    localPath=metadata["localpath"],
                    lfn=metadata["filedict"]["LFN"],
                    destinationSEList=metadata["resolvedSE"],
                    fileMetaDict=fileMetaDict,
                    masterCatalogOnly=True,
                )
                if not result["OK"]:
                    self.log.error(f"Could not transfer and register {fileName} with metadata:\n {metadata}")
                    failover[fileName] = metadata
                else:
                    # Only attempt replication after successful upload
                    lfn = metadata["lfn"]
                    uploaded.append(lfn)
                    seList = metadata["resolvedSE"]
                    replicateSE = ""
                    uploadedSE = result["Value"].get("uploadedSE", "")
                    if uploadedSE:
                        for se in seList:
                            if not se == uploadedSE:
                                replicateSE = se
                                break

                    if replicateSE and lfn and self.replicateUserOutputData:
                        self.log.info(f"Will attempt to replicate {lfn} to {replicateSE}")
                        replication[lfn] = (uploadedSE, replicateSE, fileMetaDict)

            cleanUp = False
            for fileName, metadata in failover.items():
                random.shuffle(self.failoverSEs)
                targetSE = metadata["resolvedSE"][0]
                if len(metadata["resolvedSE"]) > 1:
                    replicateSE = metadata["resolvedSE"][1]
                else:
                    replicateSE = ""
                metadata["resolvedSE"] = self.failoverSEs
                fileMetaDict = {
                    "Size": metadata["filedict"]["Size"],
                    "LFN": metadata["filedict"]["LFN"],
                    "GUID": metadata["filedict"]["GUID"],
                }
                result = self.failoverTransfer.transferAndRegisterFileFailover(
                    fileName,
                    metadata["localpath"],
                    metadata["lfn"],
                    targetSE,
                    metadata["resolvedSE"],
                    fileMetaDict=fileMetaDict,
                    masterCatalogOnly=True,
                )
                if not result["OK"]:
                    self.log.error(f"Could not transfer and register {fileName} with metadata:\n {metadata}")
                    cleanUp = True
                    continue  # for users can continue even if one completely fails
                else:
                    lfn = metadata["lfn"]
                    uploaded.append(lfn)
                    # Even when using Failover, one needs to replicate to a second SE
                    if replicateSE and self.replicateUserOutputData:
                        replication[lfn] = (targetSE, replicateSE, fileMetaDict)

            # For files correctly uploaded must report LFNs to job parameters
            if uploaded:
                report = ", ".join(uploaded)
                self.setJobParameter("UploadedOutputData", report)

            # Now after all operations, retrieve potentially modified request object
            self.request = self.failoverTransfer.request

            # If some or all of the files failed to be saved to failover
            if cleanUp:
                self.workflow_commons["Request"] = self.request
                # Leave any uploaded files just in case it is useful for the user
                # do not try to replicate any files.
                return S_ERROR("Failed To Upload Output Data")

            for lfn, (uploadedSE, repSE, fileMetaDictItem) in replication.items():
                self.failoverTransfer._setFileReplicationRequest(lfn, repSE, fileMetaDictItem, uploadedSE)

            self.workflow_commons["Request"] = self.failoverTransfer.request

            self.generateFailoverFile()

            self.setApplicationStatus("Job Finished Successfully")

            return S_OK("Output data uploaded")

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in UserJobFinalization execute module", lException=e)
            self.setApplicationStatus(repr(e))
            return S_ERROR(str(e))

        finally:
            super().finalize()

    #############################################################################

    def _getOrderedSEsList(self):
        """Returns list of ordered SEs to which trying to upload."""
        # FIXME: remove all banned SEs (not the force ones)
        # First get the local (or assigned) SE to try first for upload and others in random fashion
        localSEs = set(getDestinationSEList("Tier1-USER", self.siteName, outputmode="local"))
        self.log.verbose(f"Site Local SE for user outputs is: {list(localSEs)}")
        userSEs = set(self.userOutputSE)
        otherSEs = set(self.defaultOutputSE) - localSEs - userSEs
        # If a user SE is  local set it first
        topSEs = userSEs & localSEs
        # reordered user SEs, setting local first
        userSEs = list(topSEs) + list(userSEs - topSEs)
        localSEs = list(localSEs - topSEs)
        if len(userSEs) < 2 and localSEs:
            # Set a local SE first
            orderedSEs = localSEs[0:1] + userSEs + localSEs[1:]
        else:
            orderedSEs = userSEs + localSEs
        random.shuffle(list(otherSEs))
        orderedSEs += otherSEs

        return orderedSEs



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""ProductionRequestHandler is the implementation of the Production Request
service."""
import importlib.resources
import os
import subprocess
import tempfile

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.Core.Utilities.Subprocess import shellCall

from LHCbDIRAC.ProductionManagementSystem.DB.ProductionRequestDB import ProductionRequestDB


class ProductionRequestHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        cls.productionRequestDB = ProductionRequestDB()
        return S_OK()

    def __clientCredentials(self):
        creds = self.getRemoteCredentials()
        group = creds.get("group", "(unknown)")
        DN = creds.get("DN", "(unknown)")
        #    if 'DN' in creds:
        #      cn = re.search('/CN=([^/]+)',creds['DN'])
        #      if cn:
        #        return { 'User':cn.group(1), 'Group':group }
        return {"User": creds.get("username", "Anonymous"), "Group": group, "DN": DN}

    types_createProductionRequest = [dict]

    def export_createProductionRequest(self, requestDict):
        """Create production request."""
        creds = self.__clientCredentials()
        if "MasterID" not in requestDict:
            requestDict["RequestAuthor"] = creds["User"]
        return self.productionRequestDB.createProductionRequest(requestDict, creds)

    types_getProductionRequest = [list]

    def export_getProductionRequest(self, requestIDList):
        """Get production request(s) specified by the list of requestIDs
        AZ!!: not tested !!
        """
        if not requestIDList:
            return S_OK({})
        result = self.productionRequestDB.getProductionRequest(requestIDList)
        if not result["OK"]:
            return result
        rows = {}
        for row in result["Value"]["Rows"]:
            iD = row["RequestID"]
            rows[iD] = row
        return S_OK(rows)

    types_getProductionRequestList = [
        (int,),
        (str,),
        (str,),
        (int,),
        (int,),
        dict,
    ]

    def export_getProductionRequestList(self, subrequestFor, sortBy, sortOrder, offset, limit, rFilter):
        """Get production requests in list format (for portal grid)"""
        return self.productionRequestDB.getProductionRequest(
            [], subrequestFor, sortBy, sortOrder, offset, limit, rFilter
        )

    types_updateProductionRequest = [(int,), dict]

    def export_updateProductionRequest(self, requestID, requestDict):
        """Update production request specified by requestID."""
        creds = self.__clientCredentials()
        return self.productionRequestDB.updateProductionRequest(requestID, requestDict, creds)

    types_duplicateProductionRequest = [(int,), bool]

    def export_duplicateProductionRequest(self, requestID, clearpp):
        """Duplicate production request with subrequests."""
        creds = self.__clientCredentials()
        return self.productionRequestDB.duplicateProductionRequest(requestID, creds, clearpp)

    types_deleteProductionRequest = [(int,)]

    def export_deleteProductionRequest(self, requestID):
        """Delete production request specified by requestID."""
        creds = self.__clientCredentials()
        return self.productionRequestDB.deleteProductionRequest(requestID, creds)

    types_splitProductionRequest = [(int,), list]

    def export_splitProductionRequest(self, requestID, splitList):
        """split production request."""
        creds = self.__clientCredentials()
        return self.productionRequestDB.splitProductionRequest(requestID, splitList, creds)

    types_getProductionProgressList = [(int,)]

    def export_getProductionProgressList(self, requestID):
        """Return the list of associated with requestID productions."""
        return self.productionRequestDB.getProductionProgress(requestID)

    types_addProductionToRequest = [dict]

    def export_addProductionToRequest(self, pdict):
        """Associate production to request."""
        return self.productionRequestDB.addProductionToRequest(pdict)

    types_removeProductionFromRequest = [(int,)]

    def export_removeProductionFromRequest(self, productionID):
        """Deassociate production."""
        return self.productionRequestDB.removeProductionFromRequest(productionID)

    types_useProductionForRequest = [(int,), bool]

    def export_useProductionForRequest(self, productionID, used):
        """Set Used flags for production."""
        return self.productionRequestDB.useProductionForRequest(productionID, used)

    types_getRequestHistory = [(int,)]

    def export_getRequestHistory(self, requestID):
        """Return the list of state changes for the request."""
        return self.productionRequestDB.getRequestHistory(requestID)

    types_getTrackedProductions = []

    def export_getTrackedProductions(self):
        """Return the list of productions in active requests."""
        return self.productionRequestDB.getTrackedProductions()

    types_updateTrackedProductions = [list]

    def export_updateTrackedProductions(self, update):
        """Update tracked productions (used by Agent)"""
        return self.productionRequestDB.updateTrackedProductions(update)

    types_getTrackedInput = []

    def export_getTrackedInput(self):
        """Return the list of requests with dynamic input data."""
        return self.productionRequestDB.getTrackedInput()

    types_updateTrackedInput = [list]

    def export_updateTrackedInput(self, update):
        """Update real number of input events (used by Agent)"""
        return self.productionRequestDB.updateTrackedInput(update)

    types_getAllProductionProgress = []

    def export_getAllProductionProgress(self):
        """Return all the production progress."""
        return self.productionRequestDB.getAllProductionProgress()

    def __productionTemplatePaths(self):
        """Return production template list (file based)"""
        tplFolder = importlib.resources.files("LHCbDIRAC.ProductionManagementSystem") / "Templates"
        for path in tplFolder.iterdir():
            if not path.is_file() or path.name[-1] in [".", "~"]:
                continue
            yield path

    types_getProductionTemplateList = []

    def export_getProductionTemplateList(self):
        """Return production template list (file based)"""
        templates = [
            {
                "AuthorGroup": "",
                "Author": "",
                "PublishingTime": "",
                "LongDescription": "",
                "WFName": path.name,
                "AuthorDN": "",
                "WFParent": "",
                "Description": "",
            }
            for path in self.__productionTemplatePaths()
        ]
        return S_OK(templates)

    types_getProductionTemplate = [(str,)]

    def export_getProductionTemplate(self, name):
        for path in self.__productionTemplatePaths():
            if path.name == name:
                return S_OK(path.read_text())
        else:
            return S_ERROR(f"Template {name!r} doesn't exist")

    types_execProductionScript = [(str,), (str,)]

    def export_execProductionScript(self, script, workflow):
        creds = self.__clientCredentials()
        if creds["Group"] != "lhcb_prmgr":
            return S_ERROR("You have to be production manager")
        result = gProxyManager.downloadProxyToFile(
            creds["DN"], creds["Group"], filePath=False, requiredTimeLeft=86400, cacheTime=86400
        )
        if not result["OK"]:
            return result
        proxyFile = result["Value"]
        filesToClean = [proxyFile]

        try:
            with tempfile.NamedTemporaryFile(mode="w+t", delete=False) as workflowFile:
                filesToClean += [workflowFile.name]
                workflowFile.write(workflow)
            with tempfile.NamedTemporaryFile(mode="w+t", delete=False) as scriptFile:
                filesToClean += [scriptFile.name]
                scriptFile.write(script)

            result = subprocess.run(
                ["python", scriptFile.name, workflowFile.name],
                check=False,
                capture_output=True,
                text=True,
                timeout=1800,
                env=os.environ | {"X509_USER_PROXY": proxyFile},
            )
            if result.returncode == 0:
                return S_OK(result.stdout + result.stderr)
        finally:
            for filename in filesToClean:
                os.remove(filename)

        message = f"Command exited with: {result.returncode}\nstdout: {result.stdout}\nstderr: {result.stderr}"
        gLogger.error("Failed to execProductionScript", message)
        return S_ERROR(message)

    types_execWizardScript = [(str,), dict]

    def export_execWizardScript(self, wizard, wizpar):
        """Execure wizard with parameters."""
        creds = self.__clientCredentials()
        if creds["Group"] != "lhcb_prmgr":
            # return S_ERROR("You have to be production manager")
            if "Generate" in wizpar:
                del wizpar["Generate"]
        result = gProxyManager.downloadProxyToFile(
            creds["DN"], creds["Group"], filePath=False, requiredTimeLeft=86400, cacheTime=86400
        )
        if not result["OK"]:
            return result
        proxyFile = result["Value"]["proxyFile"]
        try:
            f = tempfile.mkstemp()
            os.write(f[0], "wizardParameters = {\n")
            for name, value in wizpar.items():
                os.write(f[0], '  "' + str(name) + '": """' + str(value) + '""",\n')
            os.write(f[0], "}\n")
            os.write(f[0], wizard)
            os.close(f[0])
        except Exception as msg:
            gLogger.error("In temporary files createion: " + str(msg))
            os.remove(proxyFile)
            return S_ERROR(str(msg))
        setenv = "source /opt/dirac/bashrc"
        # #proxy = "X509_USER_PROXY=xxx"
        proxy = f"X509_USER_PROXY={proxyFile}"
        cmd = f"python {f[1]}"
        try:
            res = shellCall(1800, [f"/bin/bash -c '{setenv};{proxy} {cmd}'"])
            if res["OK"]:
                result = S_OK(str(res["Value"][1]) + str(res["Value"][2]))
            else:
                gLogger.error(res["Message"])
                result = res
        except Exception as msg:
            gLogger.error("During execution: " + str(msg))
            result = S_ERROR(f"Failed to execute: {str(msg)}")
        os.remove(f[1])
        os.remove(proxyFile)
        return result

    types_getProductionList = [(int,)]

    def export_getProductionList(self, requestID):
        """Return the list of productions associated with request and its
        subrequests."""
        return self.productionRequestDB.getProductionList(requestID)

    types_getProductionRequestSummary = [[(str,), list], [(str,), list]]

    def export_getProductionRequestSummary(self, status, requestType):
        """Method to retrieve the production / request relations for a given
        request status."""
        if isinstance(requestType, str):
            reqTypes = [requestType]
        elif isinstance(requestType, list):
            reqTypes = requestType
        else:
            return S_ERROR(f"Invalid request type: {type(requestType)}")

        if isinstance(status, str):
            selectStatus = [status]
        elif isinstance(status, list):
            selectStatus = status
        else:
            return S_ERROR(f"Invalid status type: {type(status)}")

        res = self.productionRequestDB.getFields(
            "ProductionRequests", ["RequestID"], {"RequestState": selectStatus, "RequestType": reqTypes}
        )
        if not res["OK"]:
            return res

        reqList = self.productionRequestDB.getProductionRequest([requestID[0] for requestID in res["Value"]])
        if not reqList["OK"]:
            return reqList

        requests = reqList["Value"]
        resultDict = {}

        for req in requests["Rows"]:
            iD = int(req["RequestID"])
            if not req["RequestType"] in reqTypes:
                gLogger.verbose(f"Skipping {req['RequestType']} request ID {iD}...")
                continue
            if not req["RequestState"] in selectStatus:
                gLogger.verbose(f"Skipping request ID {iD} in state {req['RequestState']}")
                continue
            if req["HasSubrequest"]:
                gLogger.verbose(f"Simulation request {iD} is a parent, getting subrequests...")
                subReq = self.productionRequestDB.getProductionRequest([], int(iD))
                if not subReq["OK"]:
                    gLogger.error(f"Could not get production request for {iD}")
                    return subReq
                for sreq in subReq["Value"]["Rows"]:
                    sid = int(sreq["RequestID"])
                    resultDict[sid] = {"reqTotal": sreq["rqTotal"], "bkTotal": sreq["bkTotal"], "master": iD}
            else:
                gLogger.verbose(f"Simulation request {iD} is a single request")
                resultDict[iD] = {"reqTotal": req["rqTotal"], "bkTotal": req["bkTotal"], "master": 0}

        return S_OK(resultDict)

    types_getFilterOptions = []

    def export_getFilterOptions(self):
        """Return the dictionary with possible values for filter."""
        return self.productionRequestDB.getFilterOptions()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_PMS_Client_ProcessingProgress."""
import unittest

dummyResults = {}


class DummyReturn:
    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, name):
        return self.dummyMethod

    def dummyMethod(self, *args, **kwargs):
        return dummyResults[self.__class__.__name__]


class dTable(DummyReturn):
    pass


################################################################################


class HTMLProgressTable_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""

        # We need the proper software, and then we overwrite it.
        import LHCbDIRAC.ProductionManagementSystem.Client.ProcessingProgress as moduleTested

        self.progress = moduleTested.HTMLProgressTable

    def tearDown(self):
        """TearDown."""
        del self.progress


################################################################################


class HTMLProgressTable_Success(HTMLProgressTable_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""

        global dummyResults
        dummyResults["dTable"] = None

        progress = self.progress("processingPass")
        self.assertEqual("HTMLProgressTable", progress.__class__.__name__)


################################################################################



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Fix the luminosity of all descendants of a set of RAW files, if the run is finished."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()

    Script.registerSwitch("", "DoIt", "   Fix the BK database (default No)")
    Script.registerSwitch(
        "", "Force", "   Force checking all descendants and not only those of files with bad lumi (default No)"
    )
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile]",
            ]
        )
    )

    Script.addDefaultOptionValue("LogLevel", "error")
    Script.parseCommandLine(ignoreErrors=False)

    from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeFixLuminosity

    executeFixLuminosity(dmScript)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
 Test a plugin
"""
import DIRAC
from DIRAC import S_OK
from DIRAC.Core.Base.Script import Script


class FakeClient:
    def __init__(self, trans, transID, lfns, asIfProd):
        self.trans = trans
        self.transID = transID
        from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

        self.transClient = TransformationClient()
        from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

        self.bk = BookkeepingClient()
        from DIRAC.DataManagementSystem.Client.DataManager import DataManager

        self.dm = DataManager()
        self.asIfProd = asIfProd

        (self.transFiles, self.transReplicas) = self.prepareForPlugin(lfns)

    def addFilesToTransformation(self, transID, lfns):
        return S_OK({"Failed": {}, "Successful": {lfn: "Added" for lfn in lfns}})

    def getTransformation(self, transID, extraParams=False):
        if transID == self.transID and self.asIfProd:
            transID = self.asIfProd
        if transID != self.transID:
            return self.transClient.getTransformation(transID)
        res = self.trans.getType()
        return DIRAC.S_OK({"Type": res["Value"]})

    def getReplicas(self):
        return self.transReplicas

    def getFiles(self):
        return self.transFiles

    def getCounters(self, table, attrList, condDict):
        if condDict["TransformationID"] == self.transID and self.asIfProd:
            condDict["TransformationID"] = self.asIfProd
        if condDict["TransformationID"] != self.transID:
            return self.transClient.getCounters(table, attrList, condDict)
        possibleTargets = [
            "CERN-RAW",
            "CNAF-RAW",
            "GRIDKA-RAW",
            "IN2P3-RAW",
            "SARA-RAW",
            "PIC-RAW",
            "RAL-RAW",
            "NCBJ-RAW",
        ]
        counters = [({"UsedSE": se}, 0) for se in possibleTargets]
        return DIRAC.S_OK(counters)

    def getTransformations(self, cond):
        res = self.transClient.getTransformations(cond)
        if res["OK"]:
            res["Value"].append({"TransformationID": self.transID})
            if self.asIfProd and self.asIfProd not in [t["TransformationID"] for t in res["Value"]]:
                res["Value"].append({"TransformationID": self.asIfProd})
        return res

    def getBookkeepingQuery(self, transID):
        if transID == self.transID and self.asIfProd:
            return self.transClient.getBookkeepingQuery(self.asIfProd)
        return self.trans.getBkQuery()

    def insertTransformationRun(self, transID, runID, xx):
        return DIRAC.S_OK()

    def getTransformationRuns(self, condDict):
        if condDict["TransformationID"] == self.transID and self.asIfProd:
            condDict["TransformationID"] = self.asIfProd
        if condDict["TransformationID"] == self.transID:
            transRuns = []
            runs = condDict.get("RunNumber", [])
            if not runs and self.transFiles:
                res = self.bk.getFileMetadata([fileDict["LFN"] for fileDict in self.transFiles])
                if not res["OK"]:
                    return res
                runs = list({meta["RunNumber"] for meta in res["Value"]["Successful"].values()})
            for run in runs:
                transRuns.append({"RunNumber": run, "Status": "Active", "SelectedSite": None})
            return DIRAC.S_OK(transRuns)
        else:
            return self.transClient.getTransformationRuns(condDict)

    def getTransformationFiles(self, condDict=None):
        if condDict.get("TransformationID") == self.transID and self.asIfProd:
            condDict["TransformationID"] = self.asIfProd
        if condDict.get("TransformationID") == self.transID:
            transFiles = []
            if "Status" in condDict and "Unused" not in condDict["Status"]:
                return DIRAC.S_OK(transFiles)
            runs = None
            if "RunNumber" in condDict:
                runs = condDict["RunNumber"]
                if not isinstance(runs, list):
                    runs = [runs]
            for fileDict in self.transFiles:
                if not runs or fileDict["RunNumber"] in runs:
                    transFiles.append({"LFN": fileDict["LFN"], "Status": "Unused", "RunNumber": fileDict["RunNumber"]})
            return DIRAC.S_OK(transFiles)
        else:
            return self.transClient.getTransformationFiles(condDict=condDict)

    def setParameterToTransformationFiles(self, transID, lfnDict):
        """Update the transFiles with some parameters."""
        if transID == self.transID:
            for fileDict in self.transFiles:
                fileDict.update(lfnDict.get(fileDict["LFN"], {}))
            return S_OK()
        else:
            return self.transClient.setParameterToTransformationFiles(transID, lfnDict)

    def getTransformationFilesCount(self, transID, field, selection=None):
        if selection is None:
            selection = {}
        if transID == self.transID or selection.get("TransformationID") == self.transID:
            runs = selection.get("RunNumber")
            if runs and not isinstance(runs, list):
                runs = [runs]
            if field == "Status":
                counters = {"Unused": 0}
                for fileDict in self.transFiles:
                    if not runs or fileDict["RunNumber"] in runs:
                        counters["Unused"] += 1
            elif field == "RunNumber":
                counters = {}
                for fileDict in self.transFiles:
                    runID = fileDict["RunNumber"]
                    if not runs or runID in runs:
                        counters.setdefault(runID, 0)
                        counters[runID] += 1
            else:
                return DIRAC.S_ERROR("Not implemented for field " + field)
            counters["Total"] = sum(count for count in counters.values())
            return DIRAC.S_OK(counters)
        else:
            return self.transClient.getTransformationFilesCount(transID, field, selection=selection)

    def getTransformationRunStats(self, transIDs):
        counters = {}
        for transID in transIDs:
            if transID == self.transID:
                for fileDict in self.transFiles:
                    runID = fileDict["RunNumber"]
                    counters[transID][runID]["Unused"] = (
                        counters.setdefault(transID, {}).setdefault(runID, {}).setdefault("Unused", 0) + 1
                    )
                for runID in counters[transID]:
                    counters[transID][runID]["Total"] = counters[transID][runID]["Unused"]
            else:
                res = self.transClient.getTransformationRunStats(transIDs)
                if res["OK"]:
                    counters.update(res["Value"])
                else:
                    return res
        return DIRAC.S_OK(counters)

    def addRunsMetadata(self, runID, val):
        return self.transClient.addRunsMetadata(runID, val)

    def getRunsMetadata(self, runID):
        return self.transClient.getRunsMetadata(runID)

    def setTransformationRunStatus(self, transID, runID, status):
        return DIRAC.S_OK()

    def setTransformationRunsSite(self, transID, runID, site):
        return DIRAC.S_OK()

    def setFileStatusForTransformation(self, transID, status, lfns):
        return DIRAC.S_OK()

    def addTransformationRunFiles(self, transID, run, lfns):
        return DIRAC.S_OK()

    def setDestinationForRun(self, runID, site):
        return DIRAC.S_OK()

    def getDestinationForRun(self, runID):
        return self.transClient.getDestinationForRun(runID)

    def prepareForPlugin(self, lfns):
        import time

        print(f"Preparing the plugin input data ({len(lfns)} files)")
        type = self.trans.getType()["Value"]
        if not lfns:
            return (None, None)
        res = self.bk.getFileMetadata(lfns)
        if res["OK"]:
            files = []
            for lfn, metadata in res["Value"]["Successful"].items():
                runID = metadata.get("RunNumber", 0)
                runDict = {"RunNumber": runID, "LFN": lfn}
                files.append(runDict)
        else:
            print("Error getting BK metadata", res["Message"])
            return ([], {})
        replicas = {}
        startTime = time.time()
        from DIRAC.Core.Utilities.List import breakListIntoChunks

        for lfnChunk in breakListIntoChunks(lfns, 200):
            # print lfnChunk
            if type.lower() in ("replication", "removal"):
                res = self.dm.getReplicas(lfnChunk, getUrl=False)
            else:
                res = self.dm.getReplicasForJobs(lfnChunk, getUrl=False)
            # print res
            if res["OK"]:
                for lfn, ses in res["Value"]["Successful"].items():
                    if ses:
                        replicas[lfn] = sorted(ses)
            else:
                print(f"Error getting replicas of {len(lfns)} files:", res["Message"])
        print(f"Obtained replicas of {len(lfns)} files in {time.time() - startTime:.3f} seconds")
        return (files, replicas)


def printFinalSEs(transType, location, targets):
    targets = targets.split(",")
    if transType == "Removal":
        remain = []
        for x in location:
            r = ",".join([se for se in x.split(",") if se not in targets])
            remain.append(r)
        print("    Remaining SEs:", remain)
    if transType == "Replication":
        total = []
        for x in location:
            r = x + "," + ",".join([se for se in targets if se not in x.split(",")])
            total.append(r)
        print("    Final SEs:", total)


@Script()
def main():
    import DIRAC
    from LHCbDIRAC.TransformationSystem.Utilities.PluginScript import PluginScript

    pluginScript = PluginScript()
    pluginScript.registerPluginSwitches()
    pluginScript.registerFileSwitches()

    Script.registerSwitch("", "AsIfProduction=", "   Production # that this test using as source of information")
    Script.registerSwitch("", "AllFiles", "   Sets visible = False (useful if files were marked invisible)")
    Script.registerSwitch("", "NoReplicaFiles", "   Also gets the files without replica (just for BK test)")
    Script.registerSwitch("", "DefaultOptions", "   Gets from CS the default options for a plugin")

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ...",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
    from LHCbDIRAC.TransformationSystem.Client.Transformation import Transformation
    from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import getRemovalPlugins, getReplicationPlugins
    from DIRAC import gLogger

    asIfProd = None
    allFiles = False
    noRepFiles = False
    defaultOptions = False
    switches = Script.getUnprocessedSwitches()
    for opt, val in switches:
        if opt == "AsIfProduction":
            asIfProd = int(val)
        elif opt == "AllFiles":
            allFiles = True
        elif opt == "NoReplicaFiles":
            noRepFiles = True
        elif opt == "DefaultOptions":
            defaultOptions = True
    # print pluginScript.getOptions()
    plugin = pluginScript.getOption("Plugin")

    # Just get and print default options from the CS
    if defaultOptions:
        if not plugin:
            gLogger.fatal("No plugin specified")
            DIRAC.exit(1)
        # Get default options from CS
        res = Operations().getOptionsDict(f"TransformationPlugins/{plugin}")
        if res["OK"]:
            gLogger.notice(f"Parameters for plugin {plugin} (*<param> means it is a generic parameter)")
            options = res["Value"]
            # Get default options for all plugins
            res = Operations().getOptionsDict("TransformationPlugins")
            if res["OK"]:
                allOptions = res["Value"]
                for opt in set(allOptions) - set(options):
                    if opt in pluginScript.seParameters:
                        options["*" + opt] = allOptions[opt]
            # SE options first
            for opt in [opt for opt in sorted(options) if "SEs" in opt]:
                gLogger.notice(f"\t{opt} : {options[opt]}")
            # Other options
            for opt in [opt for opt in sorted(options) if "SEs" not in opt]:
                gLogger.notice(f"\t{opt} : {options[opt]}")
            DIRAC.exit(0)
        else:
            gLogger.error("Plugin has no specific parameters")
            DIRAC.exit(1)

    requestID = pluginScript.getOption("RequestID", 0)
    requestedLFNs = pluginScript.getOption("LFNs")
    # print pluginParams

    # FIXME: can be removed when the subLoggers can do it...
    gLogger.showHeaders()

    gLogger.setLevel("INFO")
    # Create the transformation
    transformation = Transformation()
    transType = ""
    if plugin in getRemovalPlugins():
        transType = "Removal"
    elif plugin in getReplicationPlugins():
        transType = "Replication"
    else:
        transType = "Processing"
    transType = pluginScript.getOption("Type", transType)
    transformation.setType(transType)

    visible = "Yes"
    if (
        allFiles
        or not plugin
        or plugin == "DestroyDataset"
        or pluginScript.getOption("Productions")
        or transType == "Processing"
    ):
        visible = "All"
    bkQueryDict = {}
    checkReplica = True
    if not requestedLFNs:
        bkQuery = pluginScript.getBKQuery(visible=visible)
        if noRepFiles and not plugin:
            # FIXME: this should work but doesn't yet...
            # bkQuery.setOption( 'ReplicaFlag', "ALL" )
            checkReplica = False
        bkQueryDict = bkQuery.getQueryDict()
        if list(bkQueryDict) in ([], ["Visible"]):
            print("No BK query was given...")
            Script.showHelp(exitCode=2)

    reqID = pluginScript.getRequestID()
    if not requestID and reqID:
        requestID = reqID

    pluginParams = pluginScript.getPluginParameters()
    pluginSEParams = pluginScript.getPluginSEParameters()
    if pluginSEParams:
        for key, val in pluginSEParams.items():
            res = transformation.setSEParam(key, val)
            if not res["OK"]:
                print(res["Message"])
                DIRAC.exit(2)
    if pluginParams:
        for key, val in pluginParams.items():
            res = transformation.setAdditionalParam(key, val)
            if not res["OK"]:
                print(res["Message"])
                DIRAC.exit(2)

    print("Transformation type:", transType)
    if requestedLFNs:
        print(f"{len(requestedLFNs)} requested LFNs")
    else:
        print("BK Query:", bkQueryDict)
    print("Plugin:", plugin)
    print("Parameters:", pluginParams)
    if pluginSEParams:
        print("SE parameters:", pluginSEParams)
    if requestID:
        print("RequestID:", requestID)
    # get the list of files from BK
    if requestedLFNs:
        lfns = requestedLFNs
    else:
        print("Getting the files from BK")
        lfns = bkQuery.getLFNs(
            printSEUsage=(
                (transType == "Removal" or not plugin)
                and not pluginScript.getOption("Runs")
                and not pluginScript.getOption("DQFlags")
            ),
            printOutput=checkReplica,
            visible=visible,
        )
        if not checkReplica:
            bkQuery.setOption("ReplicaFlag", "No")
            lfns += bkQuery.getLFNs(printSEUsage=False, printOutput=False, visible=visible)
            print(f"{len(lfns)} files in directories:")
            directories = {}
            import os

            for lfn in lfns:
                dd = os.path.dirname(lfn)
                directories[dd] = directories.setdefault(dd, 0) + 1
            for dd in sorted(directories):
                print(dd, directories[dd])
    if len(lfns) == 0:
        print("No files found in BK...Exiting now")
        DIRAC.exit(0)

    if not plugin:
        print("No plugin to be tested...")
        DIRAC.exit(0)

    print(f"\nNow testing the {transType.lower()} plugin {plugin}")
    transformation.setPlugin(plugin)
    transformation.setBkQuery(bkQueryDict)

    from LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin import TransformationPlugin

    transID = -9999
    pluginParams["TransformationID"] = transID
    pluginParams["Status"] = "Active"
    pluginParams["Type"] = transType
    # Create a fake transformation client
    fakeClient = FakeClient(transformation, transID, lfns, asIfProd)
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager

    oplugin = TransformationPlugin(
        plugin, transClient=fakeClient, dataManager=DataManager(), bkClient=BookkeepingClient()
    )
    pluginParams["TransformationID"] = transID
    pluginParams.update(pluginSEParams)
    oplugin.setParameters(pluginParams)
    replicas = fakeClient.getReplicas()
    files = fakeClient.getFiles()
    if not replicas:
        print("No replicas were found, exit...")
        DIRAC.exit(2)
    oplugin.setInputData(replicas)
    oplugin.setTransformationFiles(files)
    oplugin.setDebug(pluginScript.getOption("Debug", False))
    import time

    startTime = time.time()
    res = oplugin.run()
    print(f"Plugin took {time.time() - startTime:.1f} seconds")
    print("")
    if res["OK"]:
        print(len(res["Value"]), "tasks created")
        i = 0
        previousTask = {"First": 0, "SEs": None, "Location": None, "Tasks": 0}
        noReplicaLFNs = [lfn for _targetSE, lfnList in res["Value"] for lfn in lfnList if lfn not in replicas]
        if noReplicaLFNs:
            result = DataManager().getReplicas(noReplicaLFNs, getUrl=False)
            if res["OK"]:
                replicas.update(result["Value"]["Successful"])
        for task in res["Value"]:
            i += 1
            location = []
            for lfn in task[1]:
                ll = ",".join(sorted(replicas.get(lfn, ["Unknown"])))
                # print "LFN", lfn, l
                if ll not in location:
                    location.append(ll)
            if len(task[1]) == 1:
                # Only 1 file in task
                if previousTask["Tasks"]:
                    if task[0] == previousTask["SEs"] and location == previousTask["Location"]:
                        # Accumulate tasks for the same site and with same origin
                        previousTask["Tasks"] += 1
                        continue
                    else:
                        # Print out previous tasks
                        if previousTask["First"] == i - 1:
                            print("%d" % previousTask["First"], "- Target SEs:", previousTask["SEs"], end=" ")
                            "- 1 file - Current locations:", previousTask["Location"]
                        else:
                            print(
                                "%d:%d (%d tasks)" % (previousTask["First"], i - 1, i - previousTask["First"]), end=" "
                            )
                            "- Target SEs:", previousTask["SEs"], "- 1 file", " - Current locations:", previousTask[
                                "Location"
                            ]
                        printFinalSEs(transType, previousTask["Location"], previousTask["SEs"])
                previousTask = {"First": i, "SEs": task[0], "Location": location, "Tasks": 1}
            else:
                if previousTask["Tasks"]:
                    print("%d:%d (%d tasks)" % (previousTask["First"], i - 1, i - previousTask["First"]), end=" ")
                    "- Target SEs:", previousTask["SEs"], "- 1 file", " - Current locations:", previousTask["Location"]
                    printFinalSEs(transType, previousTask["Location"], previousTask["SEs"])
                    previousTask = {"First": 0, "SEs": None, "Location": None, "Tasks": 0}
                print(i, "- Target SEs:", task[0], f"- {len(task[1])} files", " - Current locations:", location)
                printFinalSEs(transType, location, task[0])
        if previousTask["Tasks"]:
            i += 1
            if i - previousTask["First"] == 1:
                print("%d" % previousTask["First"], "- Target SEs:", previousTask["SEs"], "- 1 file", end=" ")
                " - Current locations:", previousTask["Location"]
            else:
                print("%d:%d (%d tasks)" % (previousTask["First"], i - 1, i - previousTask["First"]), end=" ")
                "- Target SEs:", previousTask["SEs"], "- 1 file", " - Current locations:", previousTask["Location"]
            printFinalSEs(transType, previousTask["Location"], previousTask["SEs"])
    else:
        print(res["Message"])
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/python
"""
HTML.py - v0.04 2009-07-28 Philippe Lagadec

This module provides a few classes to easily generate HTML code such as tables
and lists.

Project website: http://www.decalage.info/python/html

License: CeCILL (open-source GPL compatible), see source code for details.
         http://www.cecill.info
"""

__version__ = "0.04"
__date__ = "2009-07-28"
__author__ = "Philippe Lagadec"
# -- LICENSE ------------------------------------------------------------------

# Copyright Philippe Lagadec - see http://www.decalage.info/contact for contact info
#
# This module provides a few classes to easily generate HTML tables and lists.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# A copy of the CeCILL license is also provided in these attached files:
# Licence_CeCILL_V2-en.html and Licence_CeCILL_V2-fr.html
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


# -- CHANGES ------------------------------------------------------------------

# 2008-10-06 v0.01 PL: - First version
# 2008-10-13 v0.02 PL: - added cellspacing and cellpadding to table
#                      - added functions to ease one-step creation of tables
#                        and lists
# 2009-07-21 v0.03 PL: - added column attributes and styles (first attempt)
#                        (thanks to an idea submitted by Michal Cernoevic)
# 2009-07-28 v0.04 PL: - improved column styles, workaround for Mozilla

# Table style to get thin black lines in Mozilla/Firefox instead of 3D borders
TABLE_STYLE_THINBORDER = "border: 1px solid #000000; border-collapse: collapse;"
# TABLE_STYLE_THINBORDER = "border: 1px solid #000000;"


# == CLASSES ===================================================================


class TableCell:
    """a TableCell object is used to create a cell in a HTML table. (TD or TH)

    :param text: text in the cell (may contain HTML tags). May be any object which
      can be converted to a string using str().
    :param header: bool, false for a normal data cell (TD), true for a header cell (TH)
    :param bgcolor: str, background color
    :param width: str, width
    :param align: str, horizontal alignement (left, center, right, justify or char)
    :param char: str, alignment character, decimal point if not specified
    :param charoff: str, see HTML specs
    :param valign: str, vertical alignment (top|middle|bottom|baseline)
    :param style: str, CSS style
    :param attribs: dict, additional attributes for the TD/TH tag

    Reference: http://www.w3.org/TR/html4/struct/tables.html#h-11.2.6
    """

    def __init__(
        self,
        text="",
        bgcolor=None,
        header=False,
        width=None,
        align=None,
        char=None,
        charoff=None,
        valign=None,
        style=None,
        attribs=None,
    ):
        """TableCell constructor."""
        self.text = text
        self.bgcolor = bgcolor
        self.header = header
        self.width = width
        self.align = align
        self.char = char
        self.charoff = charoff
        self.valign = valign
        self.style = style
        self.attribs = attribs
        if attribs is None:
            self.attribs = {}

    def __str__(self):
        """return the HTML code for the table cell as a string."""
        attribs_str = ""
        if self.bgcolor:
            self.attribs["bgcolor"] = self.bgcolor
        if self.width:
            self.attribs["width"] = self.width
        if self.align:
            self.attribs["align"] = self.align
        if self.char:
            self.attribs["char"] = self.char
        if self.charoff:
            self.attribs["charoff"] = self.charoff
        if self.valign:
            self.attribs["valign"] = self.valign
        if self.style:
            self.attribs["style"] = self.style
        for attr in self.attribs:
            attribs_str += f' {attr}="{self.attribs[attr]}"'
        if self.text:
            text = str(self.text)
        else:
            # An empty cell should at least contain a non-breaking space
            text = "&nbsp;"
        if self.header:
            return f"  <TH{attribs_str}>{text}</TH>\n"
        else:
            return f"  <TD{attribs_str}>{text}</TD>\n"


# ------------------------------------------------------------------------------


class TableRow:
    """a TableRow object is used to create a row in a HTML table. (TR tag)

    :param cells: list, tuple or any iterable, containing one string or TableCell
      object for each cell
    :param header: bool, true for a header row (TH), false for a normal data row (TD)
    :param bgcolor: str, background color
    :param col_align, col_valign, col_char, col_charoff, col_styles: see Table class
    :param attribs: dict, additional attributes for the TR tag

    Reference: http://www.w3.org/TR/html4/struct/tables.html#h-11.2.5
    """

    def __init__(
        self,
        cells=None,
        bgcolor=None,
        header=False,
        attribs=None,
        col_align=None,
        col_valign=None,
        col_char=None,
        col_charoff=None,
        col_styles=None,
    ):
        """TableCell constructor."""
        self.bgcolor = bgcolor
        self.cells = cells
        self.header = header
        self.col_align = col_align
        self.col_valign = col_valign
        self.col_char = col_char
        self.col_charoff = col_charoff
        self.col_styles = col_styles
        self.attribs = attribs
        if attribs is None:
            self.attribs = {}

    def __str__(self):
        """return the HTML code for the table row as a string."""
        attribs_str = ""
        if self.bgcolor:
            self.attribs["bgcolor"] = self.bgcolor
        for attr in self.attribs:
            attribs_str += f' {attr}="{self.attribs[attr]}"'
        result = f" <TR{attribs_str}>\n"
        for cell in self.cells:
            col = self.cells.index(cell)  # cell column index
            if not isinstance(cell, TableCell):
                cell = TableCell(cell, header=self.header)
            # apply column alignment if specified:
            if self.col_align and cell.align is None:
                cell.align = self.col_align[col]
            if self.col_char and cell.char is None:
                cell.char = self.col_char[col]
            if self.col_charoff and cell.charoff is None:
                cell.charoff = self.col_charoff[col]
            if self.col_valign and cell.valign is None:
                cell.valign = self.col_valign[col]
            # apply column style if specified:
            if self.col_styles and cell.style is None:
                cell.style = self.col_styles[col]
            result += str(cell)
        result += " </TR>\n"
        return result


# ------------------------------------------------------------------------------


class Table:
    """a Table object is used to create a HTML table. (TABLE tag)

    :param rows: list, tuple or any iterable, containing one iterable or TableRow
      object for each row
    :param header_row: list, tuple or any iterable, containing the header row (optional)
    :param border: str or int, border width
    :param style: str, table style in CSS syntax (thin black borders by default)
    :param width: str, width of the table on the page
    :param attribs: dict, additional attributes for the TABLE tag
    :param col_width: list or tuple defining width for each column
    :param col_align: list or tuple defining horizontal alignment for each column
    :param col_char: list or tuple defining alignment character for each column
    :param col_charoff: list or tuple defining charoff attribute for each column
    :param col_valign: list or tuple defining vertical alignment for each column
    :param col_styles: list or tuple of HTML styles for each column

    Reference: http://www.w3.org/TR/html4/struct/tables.html#h-11.2.1
    """

    def __init__(
        self,
        rows=None,
        border="1",
        style=None,
        width=None,
        cellspacing=None,
        cellpadding=4,
        attribs=None,
        header_row=None,
        col_width=None,
        col_align=None,
        col_valign=None,
        col_char=None,
        col_charoff=None,
        col_styles=None,
    ):
        """TableCell constructor."""
        self.border = border
        self.style = style
        # style for thin borders by default
        if style is None:
            self.style = TABLE_STYLE_THINBORDER
        self.width = width
        self.cellspacing = cellspacing
        self.cellpadding = cellpadding
        self.header_row = header_row
        self.rows = rows
        if not rows:
            self.rows = []
        self.attribs = attribs
        if not attribs:
            self.attribs = {}
        self.col_width = col_width
        self.col_align = col_align
        self.col_char = col_char
        self.col_charoff = col_charoff
        self.col_valign = col_valign
        self.col_styles = col_styles

    def __str__(self):
        """return the HTML code for the table as a string."""
        attribs_str = ""
        if self.border:
            self.attribs["border"] = self.border
        if self.style:
            self.attribs["style"] = self.style
        if self.width:
            self.attribs["width"] = self.width
        if self.cellspacing:
            self.attribs["cellspacing"] = self.cellspacing
        if self.cellpadding:
            self.attribs["cellpadding"] = self.cellpadding
        for attr in self.attribs:
            attribs_str += f' {attr}="{self.attribs[attr]}"'
        result = f"<TABLE{attribs_str}>\n"
        # insert column tags and attributes if specified:
        if self.col_width:
            for width in self.col_width:
                result += f'  <COL width="{width}">\n'
            # The following code would also generate column attributes for style
            # and alignement according to HTML4 specs,
            # BUT it is not supported completely (only width) on Mozilla Firefox:
            # see https://bugzilla.mozilla.org/show_bug.cgi?id=915
        # n_cols = max(len(self.col_styles), len(self.col_width),
        # len(self.col_align), len(self.col_valign))
        # for i in range(n_cols):
        #             col = ''
        # try:
        # if self.col_styles[i]:
        #                     col += ' style="%s"' % self.col_styles[i]
        # except: pass
        # try:
        # if self.col_width[i]:
        #                     col += ' width="%s"' % self.col_width[i]
        # except: pass
        # try:
        # if self.col_align[i]:
        #                     col += ' align="%s"' % self.col_align[i]
        # except: pass
        # try:
        # if self.col_valign[i]:
        #                     col += ' valign="%s"' % self.col_valign[i]
        # except: pass
        #             result += '<COL%s>\n' % col
        # First insert a header row if specified:
        if self.header_row:
            if not isinstance(self.header_row, TableRow):
                result += str(TableRow(self.header_row, header=True))
            else:
                result += str(self.header_row)
        # Then all data rows:
        for row in self.rows:
            if not isinstance(row, TableRow):
                row = TableRow(row)
            # apply column alignments  and styles to each row if specified:
            # (Mozilla bug workaround)
            if self.col_align and not row.col_align:
                row.col_align = self.col_align
            if self.col_char and not row.col_char:
                row.col_char = self.col_char
            if self.col_charoff and not row.col_charoff:
                row.col_charoff = self.col_charoff
            if self.col_valign and not row.col_valign:
                row.col_valign = self.col_valign
            if self.col_styles and not row.col_styles:
                row.col_styles = self.col_styles
            result += str(row)
        result += "</TABLE>"
        return result


# ------------------------------------------------------------------------------


class List:
    """a List object is used to create an ordered or unordered list in HTML.
    (UL/OL tag)

    Attributes:
    - lines: list, tuple or any iterable, containing one string for each line
    - ordered: bool, choice between an ordered (OL) or unordered list (UL)
    - attribs: dict, additional attributes for the OL/UL tag

    Reference: http://www.w3.org/TR/html4/struct/lists.html
    """

    def __init__(self, lines=None, ordered=False, start=None, attribs=None):
        """List constructor."""
        if lines:
            self.lines = lines
        else:
            self.lines = []
        self.ordered = ordered
        self.start = start
        if attribs:
            self.attribs = attribs
        else:
            self.attribs = {}

    def __str__(self):
        """return the HTML code for the list as a string."""
        attribs_str = ""
        if self.start:
            self.attribs["start"] = self.start
        for attr in self.attribs:
            attribs_str += f' {attr}="{self.attribs[attr]}"'
        if self.ordered:
            tag = "OL"
        else:
            tag = "UL"
        result = f"<{tag}{attribs_str}>\n"
        for line in self.lines:
            result += f" <LI>{str(line)}\n"
        result += f"</{tag}>\n"
        return result


# class Link (object):
# """
# a Link object is used to create link in HTML. (<a> tag)
#
# Attributes:
# - text: str, text of the link
# - url: str, URL of the link
# - attribs: dict, additional attributes for the A tag
#
# Reference: http://www.w3.org/TR/html4
# """
#
# def __init__(self, text, url=None, attribs=None):
#         """Link constructor"""
#         self.text = text
#         self.url = url
# if attribs:
#             self.attribs = attribs
# else:
#             self.attribs = {}
#
# def __str__(self):
#         """return the HTML code for the link as a string"""
#         attribs_str = ""
#         if self.url:  self.attribs['href'] = self.url
# for attr in self.attribs:
#             attribs_str += ' %s="%s"' % (attr, self.attribs[attr])
# return '<a%s>%s</a>' % (attribs_str, text)


# == FUNCTIONS ================================================================


# much simpler definition of a link as a function:
def Link(text, url):
    """create an reference link."""
    return f'<a href="{url}">{text}</a>'


def link(text, url):
    """create an reference link."""
    return f'<a href="{url}">{text}</a>'


def table(*args, **kwargs):
    """return HTML code for a table as a string.

    See Table class for parameters.
    """
    return str(Table(*args, **kwargs))


def list(*args, **kwargs):
    """return HTML code for a list as a string.

    See List class for parameters.
    """
    return str(List(*args, **kwargs))


# == MAIN =====================================================================

# Show sample usage when this file is launched as a script.

if __name__ == "__main__":
    # open an HTML file to show output in a browser
    f = open("test.html", "w")

    t = Table()
    t.rows.append(TableRow(["A", "B", "C"], header=True))
    t.rows.append(TableRow(["D", "E", "F"]))
    t.rows.append(("i", "j", "k"))
    f.write(str(t) + "<p>\n")
    print(str(t))
    print("-" * 79)

    t2 = Table([("1", "2"), ["3", "4"]], width="100%", header_row=("col1", "col2"), col_width=("", "75%"))
    f.write(str(t2) + "<p>\n")
    print(t2)
    print("-" * 79)

    t2.rows.append(["5", "6"])
    t2.rows[1][1] = TableCell("new", bgcolor="red")
    t2.rows.append(TableRow(["7", "8"], attribs={"align": "center"}))
    f.write(str(t2) + "<p>\n")
    print(t2)
    print("-" * 79)

    # sample table with column attributes and styles:
    table_data = [["Smith", "John", 30, 4.5], ["Carpenter", "Jack", 47, 7], ["Johnson", "Paul", 62, 10.55]]
    htmlcode = table(
        table_data,
        header_row=["Last name", "First name", "Age", "Score"],
        col_width=["", "20%", "10%", "10%"],
        col_align=["left", "center", "right", "char"],
        col_styles=["font-size: large", "", "font-size: small", "background-color:yellow"],
    )
    f.write(htmlcode + "<p>\n")
    print(htmlcode)
    print("-" * 79)

    def gen_table_squares(n):
        """Generator to create table rows for integers from 1 to n."""
        # First, header row:
        # yield TableRow(('x', 'square(x)'), header=True, bgcolor='blue')
        # Then all rows:
        for x in range(1, n + 1):
            yield (x, x * x)

    t = Table(rows=gen_table_squares(10), header_row=("x", "square(x)"))
    f.write(str(t) + "<p>\n")

    print("-" * 79)
    li = List(["aaa", "bbb", "ccc"])
    f.write(str(li) + "<p>\n")
    li.ordered = True
    f.write(str(li) + "<p>\n")
    li.start = 10
    f.write(str(li) + "<p>\n")

    f.close()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC/ResourceStatusSystem/Agent/LHCbPRProxyAgent.py."""
# First, pythonic stuff
import os
import stat

# Second, DIRAC stuff
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule


AGENT_NAME = "ResourceStatus/LHCbPRProxyAgent"


class LHCbPRProxyAgent(AgentModule):
    """Class LHCbPRProxyAgent. This agent is in charge of getting a fresh proxy
    from DIRAC. In this case, it sets the proxy as LHCbPR. See
    CS/Operations/Shift/LHCbPR.

    This is used by the LHCbPR framework, to get a new proxy every day,
    without having to use any hardcoded password ( which means, this agent will
    run on the same machine as LHCbPR ).
    """

    def initialize(self):
        """Method executed when the agent is launched. It sets the proxy as LHCbPR,
        which is stored under.

        /opt/dirac/pro/work/ResourceStatus/LHCbPRProxyAgent
        """

        try:
            self.am_setOption("shifterProxy", "LHCbPR")
            return S_OK()

        except Exception:
            errorStr = "LHCbPRProxyAgent initialization"
            self.log.exception(errorStr)
            return S_ERROR(errorStr)

    def execute(self):
        """At every execution this method will try to print the environment
        variable X509_USER_PROXY, which should contain the path to the proxy. If is
        empty, something went wrong.

        This agent should get a new proxy whenever it is close to expire.
        Appart from that, which is done internally, it is a pretty dummy
        loop.
        """

        try:
            diracCred = "/opt/dirac/pro/work/ResourceStatus/LHCbPRProxyAgent/.shifterCred"
            lhcbprCred = "/tmp/x509up_u45054"
            os.environ["X509_USER_PROXY"] = "/opt/dirac/pro/work/ResourceStatus/LHCbPRProxyAgent/.shifterCred"
            import shutil

            shutil.copyfile(diracCred, lhcbprCred)
            os.chmod(lhcbprCred, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)

            self.log.info("Loop")
            self.log.info(os.environ["X509_USER_PROXY"])
            return S_OK()

        except KeyError as x:
            self.log.exception(x)
            return S_ERROR(x)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It stores the job related information."""
from DIRAC import gLogger
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobConfiguration import JobConfiguration
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobOption import JobOption
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.File import File
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.InputFile import InputFile
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobParameters import JobParameters
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.Job import Job
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.FileParam import FileParam
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.SimulationConditions import SimulationConditions
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.DataTakingConditions import DataTakingConditions
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.FileReplica import FileReplica
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.ReplicaParam import ReplicaParam
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.Quality import Quality
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.QualityParameters import QualityParameters


class JobReader:
    """JobReader class."""

    def readJob(self, doc, fileName):
        """reads and stores the job elements from the file."""
        job = Job()
        job.fileName = fileName
        gLogger.debug("Reading job from" + str(fileName) + "XML!")

        self.__readJobConfigurations(doc, job)
        self.__readJobOptions(doc, job)
        self.__readJobTypeParameters(doc, job)
        self.__readJobInputFiles(doc, job)
        self.__readJobOutputFiles(doc, job)
        self.__readJobSimulationConditions(doc, job)
        self.__readJobDataTakingConditions(doc, job)

        gLogger.debug("Job reading finished successfully!")
        return job

    #############################################################################
    @staticmethod
    def __readJobConfigurations(doc, job):
        """reads and stores the job configurations."""

        conf = JobConfiguration()
        jobElements = doc.getElementsByTagName("Job")
        for node in jobElements:
            name = node.getAttributeNode("ConfigName")
            if name is not None:
                conf.configName = name.value
            else:
                gLogger.error("<ConfigName> XML tag is missing!!")

            version = node.getAttributeNode("ConfigVersion")
            if version is not None:
                conf.configVersion = version.value
            else:
                gLogger.error("<ConfigVersion> XML tag is missing!!")

            date = node.getAttributeNode("Date")
            if date is not None:
                conf.date = date.value
            else:
                gLogger.error("<Date> XML tag is missing!!")

            time = node.getAttributeNode("Time")
            if time is not None:
                conf.time = time.value
            else:
                gLogger.error("<Time> XML tag is missing!!")

        job.configuration = conf

    #############################################################################
    @staticmethod
    def __readJobOptions(doc, job):
        """reads and stores the job option."""
        jobOptions = doc.getElementsByTagName("JobOption")
        for node in jobOptions:
            options = JobOption()
            recipient = node.getAttributeNode("Recipient")
            if recipient is not None:
                options.recipient = recipient.value
            else:
                gLogger.warn("<Recipinet> jobOption XML tag is missing!!")

            name = node.getAttributeNode("Name")
            if name is not None:
                options.name = name.value
            else:
                gLogger.warn("<Name> JobOption XML tag is missing!!")

            value = node.getAttributeNode("Value")
            if value is not None:
                options.value = value.value
            else:
                gLogger.warn("<Value> JobOption XML tag is missing!!")

            job.options += [options]

    #############################################################################
    @staticmethod
    def __readJobTypeParameters(doc, job):
        """reads and stores the job parameters."""

        jobTypeParameters = doc.getElementsByTagName("TypedParameter")

        for node in jobTypeParameters:
            parameters = JobParameters()
            name = node.getAttributeNode("Name")
            if name is not None:
                parameters.name = name.value
            else:
                gLogger.warn("<Name> TypedParameter XML tag is missing!!")

            value = node.getAttributeNode("Value")
            if value is not None:
                parameters.value = value.value
            else:
                gLogger.warn("<Value> TypedParameter XML tag is missing!!")

            nodetype = node.getAttributeNode("Type")
            if nodetype is not None:
                parameters.type = nodetype.value
            else:
                gLogger.warn("<Type> TypedParameter XML tag is missing!!")

            job.addJobParams(parameters)

    ########################################################################
    @staticmethod
    def __readJobInputFiles(doc, job):
        """reads and stores the input files of a job."""
        jobInputFile = doc.getElementsByTagName("InputFile")
        for node in jobInputFile:
            inputFile = InputFile()
            inputFileName = node.getAttributeNode("Name")
            if inputFileName is not None:
                inputFile.name = inputFileName.value
            else:
                gLogger.error("<Name> InputFile XML tag is missing!!")

            job.addJobInputFiles(inputFile)

    ########################################################################
    def __readJobOutputFiles(self, doc, job):
        """reads the jobs output files."""
        jobOutputFiles = doc.getElementsByTagName("OutputFile")
        replicas = doc.getElementsByTagName("Replica")  # I have to check doc ? no node? !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        for node in jobOutputFiles:
            outputFile = File()

            name = node.getAttributeNode("Name")
            if name is not None:
                outputFile.name = name.value
            else:
                gLogger.error("<Name> Outputfile XML tag is missing!!")

            typeName = node.getAttributeNode("TypeName")
            if typeName is not None:
                outputFile.type = typeName.value
            else:
                gLogger.error("<Type> outputfile XML tag is missing!!")

            version = node.getAttributeNode("TypeVersion")
            if version is not None:
                outputFile.version = version.value
            else:
                gLogger.error("<Version> outputfile XML tag is missing!!")

            fileparams = node.getElementsByTagName("Parameter")
            for param in fileparams:
                outputFileParams = FileParam()

                name = param.getAttributeNode("Name")
                if name is not None:
                    outputFileParams.name = name.value
                else:
                    gLogger.error("<Name> outputfile Parameter XML tag is missing!!")

                value = param.getAttributeNode("Value")
                if value is not None:
                    outputFileParams.value = value.value
                else:
                    gLogger.error("<Value> outputfile Parameter XML tag is missing!!")

                outputFile.addFileParam(outputFileParams)
            self.__readFilereplica(replicas, outputFile)
            self.__readDataquality(node, outputFile)
            job.addJobOutputFiles(outputFile)

    ########################################################################
    @staticmethod
    def __readFilereplica(replicas, outputFile):
        """It creates the replica object and adds to the output file."""
        for replica in replicas:
            rep = FileReplica()
            param = ReplicaParam()
            name = replica.getAttributeNode("Name")
            location = replica.getAttributeNode("Location")
            if name is not None:
                param.name = name.value
            if location is not None:
                param.location = location.value
                rep.addParam(param)
            outputFile.replicas += [rep]

    ########################################################################
    @staticmethod
    def __readDataquality(node, outputFile):
        """It reads the data quality information of a given file."""
        qualities = node.getElementsByTagName("Quality")
        for quality in qualities:
            fileQuality = Quality()
            group = quality.getAttributeNode("Group")
            if group is None:
                fileQuality.group = group.value
            else:
                gLogger.warn("<Group> Quality XML tag is missing!!")

            flag = quality.getAttributeNode("Flag")
            if flag is not None:
                fileQuality.flag = flag.value
            else:
                gLogger.warn("<Flag> Quality XML tag is missing!!")

            parameters = quality.getElementsByTagName("Parameter")
            for param in parameters:
                qualityParameters = QualityParameters()
                name = param.getAttributeNode("Name")
                if name is not None:
                    qualityParameters.name = name.value
                else:
                    gLogger.warn("<Name> Quality XML tag is missing!!")

                value = param.getAttributeNode("Value")
                if value is not None:
                    qualityParameters.value = value.value
                else:
                    gLogger.warn("<Value> Quality XML tag is missing!!")

                fileQuality.addParam(qualityParameters)

            outputFile.qualities += [fileQuality]

    ########################################################################
    @staticmethod
    def __readJobSimulationConditions(doc, job):
        """reads and stores the job simulation condition."""
        gLogger.debug("Read Simulation Conditions")
        simcond = doc.getElementsByTagName("SimulationCondition")
        if not simcond:
            gLogger.debug("No Simulation Conditions:")
        elif len(simcond) > 1:
            gLogger.debug("Too many Simulation conditions:", str(simcond))
        else:
            simParam = SimulationConditions()
            node = simcond[0]
            gLogger.debug("Found Simulation conditions:", str(node))
            parameters = node.getElementsByTagName("Parameter")
            for param in parameters:
                name = param.getAttributeNode("Name")
                value = param.getAttributeNode("Value")
                if name is None or value is None:
                    gLogger.warn("<Name>  or <Value> simulation XML tag is missing!!")
                else:
                    simParam.parameters[name.value] = value.value
                    job.simulationCondition = simParam

    ########################################################################
    @staticmethod
    def __readJobDataTakingConditions(doc, job):
        """reads and stores the job data taking conditions."""
        gLogger.debug("Read DataTaking Conditions")
        daqcond = doc.getElementsByTagName("DataTakingConditions")
        if not daqcond:
            gLogger.debug("No Real DataTaking Conditions")
        elif len(daqcond) > 1:
            gLogger.debug("Too many Real DataTaking conditions:", str(daqcond))
        else:
            daqParam = DataTakingConditions()
            node = daqcond[0]
            gLogger.debug("Found Real DataTaking conditions:", str(node))
            parameters = node.getElementsByTagName("Parameter")
            for param in parameters:
                name = param.getAttributeNode("Name")
                value = param.getAttributeNode("Value")
                if name is None or value is None:
                    gLogger.warn("<Name>  or <Value> DataTakingConditions XML tag is missing!!")
                else:
                    daqParam.parameters[name.value] = value.value
                    job.dataTakingCondition = daqParam



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" ModuleBase - contains the base class for LHCb workflow modules. Defines several
    common utility methods
"""
import os
import copy
import time

from DIRAC import gLogger, siteName
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.Adler import fileAdler
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Resources.Catalog.PoolXMLFile import getGUID
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog
from DIRAC.WorkloadManagementSystem.Client.JobReport import JobReport
from DIRAC.TransformationSystem.Client.FileReport import FileReport
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

import LHCbDIRAC
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.Core.Utilities.ProductionData import getLogPath, constructProductionLFNs
from LHCbDIRAC.Workflow.Modules.ModulesUtilities import getNumberOfProcessorsToUse
from LHCbDIRAC.Core.Utilities.File import makeGuid


class ModuleBase:
    """Base class for Modules - works only within DIRAC workflows"""

    #############################################################################

    def __init__(self, loggerIn=None, operationsHelperIn=None, bkClientIn=None, dm=None):
        """Initialization of module base."""

        if loggerIn is None:
            self.log = gLogger.getSubLogger("ModuleBase")
        else:
            self.log = loggerIn

        if operationsHelperIn is None:
            self.opsH = Operations()
        else:
            self.opsH = operationsHelperIn

        if bkClientIn is None:
            self.bkClient = BookkeepingClient()
        else:
            self.bkClient = bkClientIn

        if dm is None:
            self.dataManager = DataManager()
        else:
            self.dataManager = dm

        self.requestValidator = RequestValidator()

        self.production_id = ""
        self.prod_job_id = ""
        self.jobID = 0
        self.step_number = ""
        self.step_id = ""

        self.fileReport = None
        self.jobReport = None
        self.request = None

        self.workflowStatus = None
        self.stepStatus = None
        self.workflow_commons = None
        self.step_commons = None

        self.debugSE = "CERN-DEBUG"

        self.executable = "gaudirun.py"
        self.applicationName = "Unknown"
        self.applicationVersion = "Unknown"
        self.applicationLog = ""
        self.applicationType = None
        self.usePrmon = False
        self.systemConfig = None
        self.extraPackages = None
        self.bkConfigName = None
        self.BKstepID = None
        self.condDBTag = None
        self.DDDBTag = None
        self.dqTag = None
        self.CPUe = None
        self.eventType = ""
        self.eventTimeout = None
        self.gaudiSteps = None
        self.InputData = ""
        self.inputDataList = []
        self.inputDataType = None
        self.histoName = "Hist.root"
        self.optionsFile = None
        self.optionsFormat = None
        self.optionsLine = None
        self.extraOptionsLine = None
        self.jobType = None
        self.logFilePath = None
        self.onlineCondDBTag = None
        self.onlineDDDBTag = None
        self.outputSEs = {}
        self.outputDataFileMask = None
        self.numberOfEvents = -1
        self.maxNumberOfEvents = None
        self.TCK = None
        self.mcTCK = None
        self.multicoreJob = True
        self.multicoreStep = False
        self.numberOfProcessors = 1
        self.poolXMLCatName = "pool_xml_catalog.xml"
        self.persistency = ""
        self.processingPass = None
        self.runNumber = "Unknown"
        self.simDescription = ""
        self.siteName = None
        self.stepName = None
        self.stepInputData = []
        self.XMLSummary = ""  # name of the file, not the object
        self.stepProcPass = None
        self.outputFilePrefix = ""

    @property
    def cleanedApplicationName(self):
        return self.applicationName.replace("/", "")

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Function called by all super classes."""
        self.log.info(f"===== Executing {LHCbDIRAC.version} ===== ")

        self.log.verbose(f"Executing directory for job is {os.getcwd()}")

        if production_id:
            self.production_id = production_id
        else:
            self.production_id = self.workflow_commons["PRODUCTION_ID"]  # This is a string, like '00051753'

        if prod_job_id:
            self.prod_job_id = prod_job_id
        else:
            self.prod_job_id = self.workflow_commons["JOB_ID"]

        if "JOBID" in os.environ:
            self.jobID = os.environ["JOBID"]

        if wms_job_id:
            self.jobID = wms_job_id

        if workflowStatus:
            self.workflowStatus = workflowStatus

        if stepStatus:
            self.stepStatus = stepStatus

        if wf_commons:
            self.workflow_commons = wf_commons

        if step_commons:
            self.step_commons = step_commons

        if step_number:
            self.step_number = step_number
        else:
            self.step_number = self.STEP_NUMBER  # pylint: disable=no-member

        if step_id:
            self.step_id = step_id
        else:
            self.step_id = f"{self.production_id}_{self.prod_job_id}_{self.step_number}"

        self.siteName = siteName()

    #############################################################################

    def finalize(self):
        self.log.info(f"===== Terminating {LHCbDIRAC.version} ===== ")

    #############################################################################

    def setApplicationStatus(self, status, sendFlag=True):
        """Wraps around setJobApplicationStatus of state update client."""
        if not self._WMSJob():
            return 0  # e.g. running locally prior to submission

        if self._checkWFAndStepStatus(noPrint=True):
            # The application status won't be updated in case the workflow or the step is failed already
            if not isinstance(status, str):
                status = str(status)
            self.log.verbose("setJobApplicationStatus(%d, %s)" % (self.jobID, status))
            jobStatus = self.jobReport.setApplicationStatus(status, sendFlag)
            if not jobStatus["OK"]:
                self.log.warn(jobStatus["Message"])

    #############################################################################

    def setJobParameter(self, name, value, sendFlag=True):
        """Wraps around setJobParameter of state update client."""
        if not self._WMSJob():
            return 0  # e.g. running locally prior to submission

        self.log.verbose("setJobParameter(%d,%s,%s)" % (self.jobID, name, value))

        jobParam = self.jobReport.setJobParameter(str(name), str(value), sendFlag)
        if not jobParam["OK"]:
            self.log.warn(jobParam["Message"])

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the module input parameters are resolved here."""

        self.log.verbose("workflow_commons = ", self.workflow_commons)
        self.log.verbose("step_commons = ", self.step_commons)

        self.fileReport = self._getFileReporter()
        self.jobReport = self._getJobReporter()
        self.request = self._getRequestContainer()

        self.__resolveInputWorkflow()

    #############################################################################

    def __resolveInputWorkflow(self):
        """Resolve the input variables that are in the workflow_commons."""

        self.runNumber = self.workflow_commons.get("runNumber", self.runNumber)

        self.persistency = self.workflow_commons.get("persistency", self.persistency)

        self.jobType = self.workflow_commons.get("JobType", self.jobType)

        self.poolXMLCatName = self.workflow_commons.get("poolXMLCatName", self.poolXMLCatName)

        self.InputData = self.workflow_commons.get("InputData", self.InputData)

        if "ParametricInputData" in self.workflow_commons:
            pID = copy.deepcopy(self.workflow_commons["ParametricInputData"])
            if pID:
                if isinstance(pID, list):
                    pID = ";".join(pID)
                #      self.InputData += ';' + pID
                self.InputData = pID
                self.InputData = self.InputData.rstrip(";")

        if self.InputData == ";":
            self.InputData = ""

        self.inputDataList = [lfn.strip("LFN:") for lfn in self.InputData.split(";") if lfn]

        # only required until the stripping is the same for MC / data
        self.bkConfigName = self.workflow_commons.get("configName", self.bkConfigName)

        self.simDescription = self.workflow_commons.get("simDescription", self.simDescription)

        if "runMetadata" in self.workflow_commons:
            runMetadataDict = eval(self.workflow_commons["runMetadata"])
            self.onlineDDDBTag = runMetadataDict.get("DDDB")
            self.onlineCondDBTag = runMetadataDict.get("CondDb")
            self.TCK = runMetadataDict.get("TCK")
            if not (self.onlineDDDBTag and self.onlineCondDBTag and self.TCK):
                self.log.error(
                    "Missing normally required Run Metadata",
                    f"DDDB={self.onlineDDDBTag}, CondDb={self.onlineCondDBTag}, TCK={self.TCK}",
                )

        if "outputDataFileMask" in self.workflow_commons:
            self.outputDataFileMask = self.workflow_commons["outputDataFileMask"]
            if not isinstance(self.outputDataFileMask, list):
                self.outputDataFileMask = [i.lower().strip() for i in self.outputDataFileMask.split(";")]

        self.gaudiSteps = self.workflow_commons.get("gaudiSteps", self.gaudiSteps)

        if "CPUe" in self.workflow_commons:
            self.CPUe = int(round(float(self.workflow_commons["CPUe"])))

        multicoreJob = self.workflow_commons.get("multicore", self.multicoreJob)  # Flag specific to production jobs
        if isinstance(multicoreJob, bool):
            self.multicoreJob = multicoreJob
        else:
            self.multicoreJob = False
            if isinstance(multicoreJob, str) and multicoreJob.lower() in ("true", "y", "yes"):
                self.multicoreJob = True

        self.usePrmon = self.opsH.getValue("Productions/UploadES_PrMon", False)

        self.processingPass = self.workflow_commons.get("processingPass", self.processingPass)

        self.logFilePath = self.workflow_commons.get("LogFilePath", self.logFilePath)
        if isinstance(self.logFilePath, list):
            self.logFilePath = self.logFilePath[0]
        else:
            if "PRODUCTION_ID" and "JOB_ID" and "configVersion" and "configName" in self.workflow_commons:
                self.log.info("LogFilePath parameter not found, creating on the fly")
                result = getLogPath(self.workflow_commons, self.bkClient)
                if not result["OK"]:
                    self.log.error("Could not create LogFilePath", result["Message"])
                    raise RuntimeError(result["Message"])
                self.logFilePath = result["Value"]["LogFilePath"][0]

        if "maxNumberOfEvents" in self.workflow_commons:
            self.maxNumberOfEvents = int(self.workflow_commons["maxNumberOfEvents"])

        self.eventType = self.workflow_commons.get("eventType", self.eventType)

        self.numberOfEvents = int(self.workflow_commons.get("numberOfEvents", self.numberOfEvents))

        if "outputSEs" in self.workflow_commons:
            self.outputSEs = self.workflow_commons["outputSEs"]
        else:
            # this is here for backward compatibility
            histogramSE = self.opsH.getValue("Productions/HistogramSE", "CERN-HIST")
            histoTypes = self.opsH.getValue(
                "Productions/HistogramTypes", ["HIST", "BRUNELHIST", "DAVINCIHIST", "GAUSSHIST"]
            )
            self.outputSEs = {ht: histogramSE for ht in histoTypes}
        # for older productions we construct it based on what should be found in the steps
        if "listoutput" in self.step_commons:
            listOutputStep = self.step_commons["listoutput"]
            for lOutput in listOutputStep:
                try:
                    for outputDataType in lOutput["outputDataType"].split(";"):
                        if outputDataType:
                            self.outputSEs.setdefault(outputDataType.upper(), lOutput["outputDataSE"])
                except KeyError:
                    continue
                self.workflow_commons["outputSEs"] = self.outputSEs

    #############################################################################

    def _resolveInputStep(self):
        """Resolve the input variables for an application step."""

        prodID = self.workflow_commons.get("PRODUCTION_ID", "")
        jobID = self.workflow_commons.get("JOB_ID", "")
        stepInstanceNumber = self.step_commons.get("STEP_NUMBER", "")

        self.stepName = self.step_commons["STEP_INSTANCE_NAME"]

        self.executable = self.step_commons.get("executable", self.executable)

        self.applicationName = self.step_commons.get("applicationName", self.applicationName)

        self.applicationVersion = self.step_commons.get("applicationVersion", self.applicationVersion)

        self.BKstepID = self.step_commons.get("BKStepID", self.BKstepID)

        self.stepProcPass = self.step_commons.get("StepProcPass", self.stepProcPass)

        if self.step_commons.get("eventTimeout"):
            self.eventTimeout = int(self.step_commons["eventTimeout"])

        # this is only for production jobs and for application steps
        if prodID and jobID and stepInstanceNumber and "listoutput" in self.step_commons:
            self.outputFilePrefix = f"{prodID}_{jobID}_{stepInstanceNumber}"
            self.applicationLog = self.cleanedApplicationName + "_" + self.outputFilePrefix + ".log"
            self.XMLSummary = "summary" + self.cleanedApplicationName + "_" + self.outputFilePrefix + ".xml"
            self.histoName = self.cleanedApplicationName + "_" + self.outputFilePrefix + ".Hist.root"

            for fileTypeDict in self.step_commons["listoutput"]:  # this is a dict like {'outputDataType': 'sim'}
                # for non histo-merging prods
                if "hist" in fileTypeDict["outputDataType"].lower() and self.jobType.lower() != "merge":
                    # Watch out: this assumes that:
                    # - 'hist' is always in the file type name
                    # - merging jobs won't produce histograms
                    # - the only merging jobs that produce output types with hist are histomerging productions
                    if "outputDataName" not in fileTypeDict:
                        fileTypeDict["outputDataName"] = self.histoName
                else:
                    fileTypeDict["outputDataName"] = self.outputFilePrefix + "." + fileTypeDict["outputDataType"]
        else:
            self.applicationLog = self.step_commons.get("applicationLog", self.applicationLog)

        self.inputDataType = self.step_commons.get("inputDataType", self.inputDataType)

        self.applicationType = self.step_commons.get("applicationType", self.applicationType)

        self.optionsFile = self.step_commons.get("optionsFile", self.optionsFile)

        self.optionsLine = self.step_commons.get("optionsLine", self.optionsLine)

        self.extraOptionsLine = self.step_commons.get("extraOptionsLine", self.extraOptionsLine)

        if "extraPackages" in self.step_commons:
            self.extraPackages = self.step_commons["extraPackages"]
            if self.extraPackages:
                if isinstance(self.extraPackages, str):
                    eps = self.extraPackages.split(";")  # pylint: disable=no-member
                    epList = []
                    for ep in eps:
                        epList.append(tuple(ep.split(".")))
                    self.extraPackages = epList

        stepInputData = []
        if "inputData" in self.step_commons:
            if self.step_commons["inputData"]:
                stepInputData = self.step_commons["inputData"]
        elif self.InputData:
            stepInputData = copy.deepcopy(self.InputData)
        if stepInputData:
            stepInputData = self._determineStepInputData(
                stepInputData,
            )
            self.stepInputData = [sid.strip("LFN:") for sid in stepInputData]

        self.optionsFormat = self.step_commons.get("optionsFormat", self.optionsFormat)

        multicoreStep = self.step_commons.get("multiCore", self.multicoreStep)
        if isinstance(multicoreStep, bool):
            self.multicoreStep = multicoreStep
        else:
            self.multicoreStep = False
            if isinstance(multicoreStep, str) and multicoreStep.lower() in ("true", "y", "yes"):
                self.multicoreStep = True

        if (self.multicoreJob and self.multicoreStep) or (
            self.jobType.lower() == "user" and self.workflow_commons.get("MaxNumberOfProcessors", False)
        ):
            # the parameter 'MaxNumberOfProcessors' is in the workflow commons:
            # for simplicity we assume that there's a general max number
            # of processors the application can use, set at the workflow level (by the Job API)
            payloadProcessors = self.workflow_commons.get("MaxNumberOfProcessors", None)
            self.numberOfProcessors = getNumberOfProcessorsToUse(jobID, payloadProcessors)

        self.systemConfig = self.step_commons.get("SystemConfig", self.systemConfig)

        self.mcTCK = self.step_commons.get("mcTCK", self.mcTCK)

        self.DDDBTag = self.step_commons.get("DDDBTag", self.DDDBTag)

        self.condDBTag = self.step_commons.get("CondDBTag", self.condDBTag)

        self.dqTag = self.step_commons.get("DQTag", self.dqTag)

        # This is resolved also in __resolveInputWorkflow but can be specialized per step (e.g. LHCbJob().setApplication())
        self.numberOfEvents = int(self.step_commons.get("numberOfEvents", self.numberOfEvents))

    #############################################################################

    def _determineOutputs(self):
        """Method that determines the correct outputs.

        For merging jobs the output has normally to be the same as the input, but there might be exceptions
        (like for the productions for the merging of histograms)
        For the others, we use what is in the step definition

        We always remove the 'HIST'(s), when present, from the list of output file types
        as these are treated differently.

        There is  anyway also here the special case of histogram merging productions.
        """

        histoTypes = self.opsH.getValue(
            "Productions/HistogramTypes", ["HIST", "BRUNELHIST", "DAVINCIHIST", "GAUSSHIST"]
        )

        stepOutputs = self.step_commons["listoutput"]
        stepOutputsT = [x["outputDataType"] for x in stepOutputs]
        stepOutTypes = []
        for fts in stepOutputsT:
            for ft in fts.split(";"):
                if ft and ft not in stepOutTypes:
                    stepOutTypes.append(ft.lower())

        # corrections for Merge productions
        if self.jobType.lower() in ("merge", "histomerge"):
            if len(stepOutTypes) == 1 and stepOutTypes[0] in [hts.lower() for hts in histoTypes] + ["root"]:
                # If it is root/histo/ntuple merging job, we treat it almost as a stripping job
                pass
            else:
                res = self.bkClient.getFileMetadata(self.stepInputData)
                if not res["OK"]:
                    raise RuntimeError(res["Message"])

                outputTypes = set()
                success = res["Value"]["Successful"]
                for mdDict in success.values():
                    outputTypes.add(mdDict["FileType"])
                if len(success) != len(self.stepInputData):
                    self.log.warn("Some inputs are not in BKK, trying to parse the file names")
                    for sid in set(self.stepInputData) - set(success):
                        # File types in the BK are upper case
                        fType = ".".join(os.path.basename(sid).split(".")[1:]).upper()
                        outputTypes.add(fType)

                outputTypes = list(outputTypes)
                if len(outputTypes) > 1:
                    raise ValueError(f"Not all input files have the same type ({','.join(outputTypes)})")
                outputType = outputTypes[0].lower()
                stepOutTypes = [outputType.lower()]
                stepOutputs = [
                    {
                        "outputDataName": self.step_id + "." + outputType.lower(),
                        "outputDataType": outputType.lower(),
                        "outputBKType": outputType.upper(),
                    }
                ]

        histogram = False

        # first here treating the special case of root/histo/ntuple merging jobs
        if (
            self.jobType.lower() in ("merge", "histomerge")
            and len(stepOutTypes) == 1
            and stepOutTypes[0] in [hts.lower() for hts in histoTypes] + ["root"]
        ):
            pass
        else:  # This is not a histogram merging production, so we remove the histograms from the step outputs
            for hist in histoTypes:
                try:
                    stepOutTypes.remove(hist)
                    histogram = True
                except ValueError:
                    pass
                try:
                    stepOutTypes.remove(hist.lower())
                    histogram = True
                except ValueError:
                    pass

        return stepOutputs, stepOutTypes, histogram

    #############################################################################

    def _getJobReporter(self):
        """just return the job reporter (object, always defined by dirac-
        jobexec)"""

        if "JobReport" in self.workflow_commons:
            return self.workflow_commons["JobReport"]
        jobReport = JobReport(self.jobID)
        self.workflow_commons["JobReport"] = jobReport
        return jobReport

    #############################################################################

    def _getFileReporter(self):
        """just return the file reporter (object)"""

        if "FileReport" in self.workflow_commons:
            return self.workflow_commons["FileReport"]
        fileReport = FileReport()
        self.workflow_commons["FileReport"] = fileReport
        return fileReport

    #############################################################################

    def _getRequestContainer(self):
        """just return the Request reporter (object)"""

        if "Request" in self.workflow_commons:
            return self.workflow_commons["Request"]
        request = Request()
        self.workflow_commons["Request"] = request
        return request

    #############################################################################

    def getCandidateFiles(self, outputList, outputLFNs, fileMask="", stepMask=""):
        """Returns list of candidate files to upload, check if some outputs are
        missing.

        :param list outputList: list of outputs with the following structure::
            [{'outputDataType': '', 'outputDataName': ''} , {...}]
        :param list outputLFNs: output LFNs for the job
        :param str fileMask: the output file extensions to restrict the outputs to. Can also be a list of strings
        :param str stepMask: the step ID to restrict the outputs to. Can also be a list of strings.

        :returns: dictionary containing type, SE and LFN for files restricted by mask
        """
        fileInfo = {}

        for outputFile in outputList:
            if "outputDataType" in outputFile and "outputDataName" in outputFile:
                fname = outputFile["outputDataName"]
                fileType = outputFile["outputDataType"]
                fileInfo[fname] = {"type": fileType}
            else:
                self.log.error("Ignoring mal-formed output data specification", str(outputFile))

        for lfn in outputLFNs:
            if os.path.basename(lfn).lower() in list(fi.lower() for fi in fileInfo):
                try:
                    fileInfo[os.path.basename(lfn)]["lfn"] = lfn
                    self.log.verbose("Found LFN for file", f"{lfn} -> {os.path.basename(lfn)}")
                except KeyError:
                    fileInfo[os.path.basename(lfn).lower()]["lfn"] = lfn
                    self.log.verbose("Found LFN for file", f"{lfn} -> {os.path.basename(lfn).lower()}")
            elif os.path.basename(lfn).split("_")[-1].lower() in list(fi.lower() for fi in fileInfo):
                try:
                    fileInfo[os.path.basename(lfn).split("_")[-1]]["lfn"] = lfn
                    self.log.verbose("Found LFN for file", f" {lfn} -> {os.path.basename(lfn).split('_')[-1]}")
                except KeyError:
                    fileInfo[os.path.basename(lfn).split("_")[-1].lower()]["lfn"] = lfn
                    self.log.verbose("Found LFN for file", f" {lfn} -> {os.path.basename(lfn).split('_')[-1].lower()}")
            else:
                self.log.warn("LFN not recognized", f"for LFN {lfn}")

        # check local existance
        fileList = self._checkLocalExistance(list(fileInfo))
        # really horrible stuff for updating the name with what's found on the disk
        # (because maybe the case is not the same as the expected)
        newFileInfo = {}
        for fi in fileInfo.items():
            for li in fileList:
                if fi[0].lower() == li.lower():
                    newFileInfo[li] = fi[1]

        # Select which files have to be uploaded: in principle all
        candidateFiles = self._applyMask(newFileInfo, fileMask, stepMask)

        # Sanity check all final candidate metadata keys are present
        self._checkSanity(candidateFiles)

        # Adding the SEs
        for candidateFile in candidateFiles:
            try:
                fType = candidateFiles[candidateFile]["type"]
                for fType in [fType, fType.lower(), fType.upper(), fType.capitalize()]:
                    try:
                        wfSE = self.outputSEs[fType]
                        candidateFiles[candidateFile]["workflowSE"] = wfSE
                        break
                    except KeyError:
                        continue
            except AttributeError:
                break

        return candidateFiles

    #############################################################################

    def _checkLocalExistance(self, fileList):
        """Check that the list of output files are present locally."""

        notPresentFiles = []

        filesOnDisk = set(os.listdir("."))
        diff = set(fileList).difference(filesOnDisk)
        if diff:
            self.log.warn("Not all files found", f"set of what's on the disk: {filesOnDisk}")
            self.log.warn("Looking for files with different case")
            diffCI = {fx.lower() for fx in fileList}.difference({fod.lower() for fod in filesOnDisk})
            if diffCI:
                self.log.error("Output data not found", f"File list {notPresentFiles} does not exist locally")
                raise OSError("Output data not found")

            # now checking what's the actual filename case that's written
            self.log.warn("Found files with filename with different case, returning those")
            filesActuallyOnDisk = []
            for fod in os.listdir("."):
                if fod.lower() in [fl.lower() for fl in fileList]:
                    filesActuallyOnDisk.append(fod)
            return filesActuallyOnDisk

        return fileList

    #############################################################################

    def _applyMask(self, candidateFilesIn, fileMask, stepMask):
        """Select which files have to be uploaded: in principle all.

        :param dict candidateFilesIn: dictionary like
          {'00012345_00012345_4.dst': {'lfn': '/lhcb/MC/2010/DST/123/123_45_4.dst',
                                       type': 'dst'},
           '00012345_00012345_2.digi': {'type': 'digi'}}
        :param str fileMask: the output file extensions to restrict the outputs to. Can also be a list of strings
        :param str stepMask: the step ID to restrict the outputs to. Can also be a list of strings.

        :returns: a dict like the one in candidateFilesIn
        """
        candidateFiles = copy.deepcopy(candidateFilesIn)

        if fileMask and not isinstance(fileMask, list):
            fileMask = [fileMask]
        if isinstance(stepMask, int):
            stepMask = str(stepMask)
        if stepMask and not isinstance(stepMask, list):
            stepMask = [stepMask]

        if fileMask and fileMask != [""]:
            for fileName, metadata in list(candidateFiles.items()):
                if metadata["type"].lower() not in [fm.lower() for fm in fileMask]:
                    del candidateFiles[fileName]
                    self.log.info(
                        "Output file %s was produced but will not be treated (fileMask is %s)"
                        % (fileName, ", ".join(fileMask))
                    )
        else:
            self.log.info("No outputDataFileMask provided, the files with all the extensions will be considered")

        if stepMask and stepMask != [""]:
            for fileName, metadata in list(candidateFiles.items()):
                if fileName.lower().replace(metadata["type"].lower(), "").split("_")[-1].split(".")[0] not in stepMask:
                    del candidateFiles[fileName]
                    self.log.info(
                        "Output file %s was produced but will not be treated (stepMask is %s)"
                        % (fileName, ", ".join(stepMask))
                    )
        else:
            self.log.info("No outputDataStep provided, the files output of all the steps will be considered")

        return candidateFiles

    #############################################################################

    def _checkSanity(self, candidateFiles):
        """Sanity check all final candidate metadata keys are present.

        :param dict candidateFiles: dictionary like
          {'00012345_00012345_4.dst': {'lfn': '/lhcb/MC/2010/DST/123/123_45_4.dst',
                                       type': 'dst'},
           '00012345_00012345_2.digi': {'type': 'digi'}}

        :returns: None or raises ValueError
        """

        notPresentKeys = []

        mandatoryKeys = ["type", "lfn"]  # filedict is used for requests
        for fileName, metadata in candidateFiles.items():
            for key in mandatoryKeys:
                if key not in metadata:
                    notPresentKeys.append((fileName, key))

        if notPresentKeys:
            for fileName_keys in notPresentKeys:
                self.log.error(f"File {fileName_keys[0]} has missing {fileName_keys[1]}")
            raise ValueError("Missing requested fileName keys")

    #############################################################################

    def getFileMetadata(self, candidateFiles):
        """Returns the candidate file dictionary with associated metadata.

        The input candidate files dictionary has the structure:

        .. code-block:: python

          {
            'foo_1.txt': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
                          'type': 'txt',
                          'workflowSE': SE1},
            'bar_2.py': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                         'type': 'py',
                         'workflowSE': 'SE2'},
          }

        this also assumes the files are in the current working directory.
        """
        # Retrieve the POOL File GUID(s) for any final output files
        self.log.info(f"Will search for POOL GUIDs for: {', '.join(candidateFiles)}")
        pfnGUID = getGUID(list(candidateFiles))
        if not pfnGUID["OK"]:
            self.log.error(
                """PoolXMLFile failed to determine POOL GUID(s) for output file list,
      these will be generated by the DataManager""",
                pfnGUID["Message"],
            )
            for fileName in candidateFiles:
                candidateFiles[fileName]["guid"] = ""
        elif pfnGUID["generated"]:
            self.log.warn("PoolXMLFile generated GUID(s) for the following files ", ", ".join(pfnGUID["generated"]))
        else:
            self.log.info(f"GUIDs found for all specified POOL files: {', '.join(candidateFiles)}")

        for pfn, guid in pfnGUID["Value"].items():
            candidateFiles[pfn]["guid"] = guid

        # Get all additional metadata about the file necessary for requests
        final = {}
        for fileName, metadata in candidateFiles.items():
            fileDict = {}
            fileDict["LFN"] = metadata["lfn"]
            fileDict["Size"] = os.path.getsize(fileName)
            fileDict["Checksum"] = fileAdler(fileName)
            fileDict["ChecksumType"] = "ADLER32"
            fileDict["GUID"] = metadata["guid"]
            fileDict["Status"] = "Waiting"

            final[fileName] = metadata
            final[fileName]["filedict"] = fileDict
            final[fileName]["localpath"] = f"{os.getcwd()}/{fileName}"

        # Sanity check all final candidate metadata keys are present (return S_ERROR if not)
        mandatoryKeys = ["guid", "filedict"]  # filedict is used for requests (this method adds guid and filedict)
        for fileName, metadata in final.items():
            for key in mandatoryKeys:
                if key not in metadata:
                    raise RuntimeError(f"File {fileName} has missing {key}")

        return final

    #############################################################################

    def _determineStepInputData(self, inputData):
        """determine the input data for the step."""
        if inputData == "previousStep":
            stepIndex = self.gaudiSteps.index(self.stepName)
            stepInputData = []
            for previousStep in self.gaudiSteps[stepIndex - 1 :: -1]:
                outputList = self.workflow_commons.get("outputList", [])
                if not outputList:
                    raise RuntimeError(
                        f"outputList was empty while finding input data for step {self.stepName} - "
                        "did the first step produce any output at all?"
                    )

                for outputF in outputList:
                    try:
                        if (
                            outputF["stepName"] == previousStep
                            and outputF["outputBKType"].lower() == self.inputDataType.lower()
                        ):
                            stepInputData.append(outputF["outputDataName"])
                    except KeyError:
                        raise RuntimeError(f"Can't find output of step {previousStep}")

                # outputDataName is always lower case but the job output can vary
                # Fix it if the file only exists in the current directory with mixed case
                filenameMap = {fn.lower(): fn for fn in os.listdir(".")}
                stepInputData = [fn if fn in os.listdir(".") else filenameMap.get(fn, fn) for fn in stepInputData]

                # We found the output file so no need to look at earlier steps
                if stepInputData:
                    break
            return stepInputData

        return [x.strip("LFN:") for x in inputData.split(";")]

    #############################################################################

    def _manageAppOutput(self, outputs):
        """Calls self._findOutputs to find what's produced, then creates the LFNs.

        outputs, as called here, is created starting from step_commons['listoutput'],
        but enriched with at least the outputDataName.

        example of outputs:
        [{'outputDataType': 'bhadron.dst', 'outputBKType': 'BHADRON.DST',
          'outputDataName': '00012345_00012345_2.BHADRON.DST'},
         {'outputDataType': 'calibration.dst','outputDataType': 'CALIBRATION.DST',
          'outputDataName': '00012345_00012345_2.CALIBRATION.DST'}]

        :params list outputs: list of dicts of step output files descriptions
        """

        if not outputs:
            self.log.warn("Step outputs are not defined (normal user jobs. Not normal in productions and SAM jobs)")
            return
        else:
            finalOutputs, _bkFileTypes = self._findOutputs(outputs)

        self.log.info(f"Final step outputs are: {finalOutputs}")
        self.step_commons["listoutput"] = finalOutputs

        if "outputList" in self.workflow_commons:
            for outFile in finalOutputs:
                if outFile not in self.workflow_commons["outputList"]:
                    self.workflow_commons["outputList"].append(outFile)
        else:
            self.workflow_commons["outputList"] = finalOutputs

        if "PRODUCTION_ID" and "JOB_ID" and "configVersion" and "configName" in self.workflow_commons:
            self.log.info("Attempting to recreate the production output LFNs...")
            result = constructProductionLFNs(self.workflow_commons, self.bkClient)
            if not result["OK"]:
                raise OSError(f"Could not create production LFNs: {result['Message']}")
            self.workflow_commons["BookkeepingLFNs"] = result["Value"]["BookkeepingLFNs"]
            self.workflow_commons["LogFilePath"] = result["Value"]["LogFilePath"]
            self.workflow_commons["ProductionOutputData"] = result["Value"]["ProductionOutputData"]

    #############################################################################

    def _findOutputs(self, stepOutput):
        """Find which outputs of those in stepOutput (what are expected to be
        produced) are effectively produced. stepOutput, as called here, is created
        starting from step_commons['listoutput']

        example of stepOutput:
        [{'outputDataType': 'bhadron.dst', 'outputBKType': 'BHADRON.DST',
          'outputDataName': '00012345_00012345_2.BHADRON.DST'},
         {'outputDataType': 'calibration.dst','outputDataType': 'CALIBRATION.DST',
          'outputDataName': '00012345_00012345_2.CALIBRATION.DST'}]

        :params list stepOutput: list of dicts of step output files descriptions
        :returns: list, list
        """

        bkFileTypes = []  # uppercase list of file types
        finalOutputs = []  # list of dicts of what's found on the local disk

        filesFound = []

        for output in stepOutput:
            found = False
            fileOnDisk = None

            for fileOnDisk in os.listdir("."):
                if output["outputDataName"].lower() == fileOnDisk.lower():
                    found = True
                    break

            if found and fileOnDisk:
                self.log.info(
                    "Found output file (case is not considered)",
                    f"{fileOnDisk} matching {output['outputDataName']} ",
                )
                output["outputDataName"] = fileOnDisk
                filesFound.append(output)
            else:
                self.log.error("Output not found", output["outputDataName"])
                raise OSError("OutputData not found")

        for fileFound in filesFound:
            bkFileTypes.append(fileFound["outputDataType"].upper())
            finalOutputs.append(
                {
                    "outputDataName": fileFound["outputDataName"],
                    "outputDataType": fileFound["outputDataType"].lower(),
                    "outputBKType": fileFound["outputDataType"].upper(),
                    "stepName": self.stepName,
                }
            )
            try:
                physFileName = fileFound["outputDataName"]
                poolCat = PoolXMLCatalog(self.poolXMLCatName)
                if not poolCat.getTypeByPfn(physFileName):
                    self.log.info("Output file not found in pool catalog. Will add it.", physFileName)
                    poolCat.addFile(
                        (
                            physFileName,
                            physFileName,
                            None,
                            makeGuid(physFileName)[physFileName],
                            "ROOT_All",
                        )
                    )
                    poolCat.flush()
                else:
                    self.log.info("Output file exists already in pool catalog. Won't touch.", physFileName)
            except Exception:
                self.log.exception(
                    "Couldn't add missing output file to pool catalog",
                    fileFound,
                )
        return (finalOutputs, bkFileTypes)

    #############################################################################

    def _WMSJob(self):
        """Check if this job is running via WMS."""
        return bool(self.jobID)

    #############################################################################

    def _enableModule(self):
        """Enable module if it's running via WMS."""
        if not self._WMSJob():
            self.log.info("No WMS JobID found, disabling module via control flag")
            return False
        self.log.verbose("Found WMS JobID", self.jobID)
        return True

    #############################################################################

    def _checkWFAndStepStatus(self, noPrint=False):
        """Check the WF and Step status."""
        if not self.workflowStatus["OK"] or not self.stepStatus["OK"]:
            if not noPrint:
                self.log.info("Skip this module, failure detected in a previous step")
                self.log.info("Workflow status", self.workflowStatus)
                self.log.info("Step Status", self.stepStatus)
            return False
        else:
            return True

    #############################################################################

    def _disableWatchdogCPUCheck(self):
        """just writes a file to disable the watchdog."""
        self.log.info("Creating DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK in order to disable the Watchdog")
        with open("DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK", "w") as fopen:
            fopen.write(f"{time.asctime()}")

    #############################################################################

    def generateFailoverFile(self):
        """Retrieve the accumulated reporting request, and produce a JSON file that
        is consumed by the JobWrapper."""
        reportRequest = None
        result = self.jobReport.generateForwardDISET()
        if not result["OK"]:
            self.log.warn("Could not generate Operation for job report", f"with result:\n{result}")
        else:
            reportRequest = result["Value"]
        if reportRequest:
            self.log.info("Populating request with job report information")
            self.request.addOperation(reportRequest)

        if len(self.request):
            try:
                optimized = self.request.optimize()
            except AttributeError:
                optimized = {"OK": True}
            if not optimized["OK"]:
                self.log.error("Could not optimize", optimized["Message"])
                self.log.error("Not failing the job because of that, keep going")
            isValid = self.requestValidator.validate(self.request)
            if not isValid["OK"]:
                raise RuntimeError(f"Failover request is not valid: {isValid['Message']}")
            else:
                requestJSON = self.request.toJSON()
                if requestJSON["OK"]:
                    self.log.info("Creating failover request for deferred operations", "for job %d" % self.jobID)
                    request_string = str(requestJSON["Value"])
                    self.log.debug(request_string)
                    # Write out the request string
                    fname = f"{self.production_id}_{self.prod_job_id}_request.json"
                    with open(fname, "w") as jsonFile:
                        jsonFile.write(request_string)
                    self.log.info("Created file containing failover request", fname)
                    result = self.request.getDigest()
                    if result["OK"]:
                        self.log.info("Digest of the request", result["Value"])
                    else:
                        self.log.warn("No digest? That's not sooo important", f"anyway: {result['Message']}")
                else:
                    raise RuntimeError(requestJSON["Message"])

        accountingReport = None
        if "AccountingReport" in self.workflow_commons:
            accountingReport = self.workflow_commons["AccountingReport"]
        if accountingReport:
            result = accountingReport.commit()
            if not result["OK"]:
                self.log.error("!!! Both accounting and RequestDB are down? !!!")
                self.log.error(f"accountingReport result: {result['Message']}")
                self.log.error(
                    'Anyway, the job won\'t fail for this reason, because this is "just" the accounting report'
                )

    #############################################################################

    def setBKRegistrationRequest(
        self, lfn, error="", metaData={"Checksum": "justSomething", "ChecksumType": "ADLER32", "GUID": "aGUID"}
    ):
        """Set a BK registration request for changing the replica flag.

        Uses the global request object (self.request).
        """
        if error:
            self.log.info(f'BK registration for {lfn} failed with message: "{error}" setting failover request')
        else:
            self.log.info(f"Setting BK registration request for {lfn}")

        regFile = Operation()
        regFile.Type = "RegisterFile"
        regFile.Catalog = "BookkeepingDB"

        bkFile = File()
        bkFile.LFN = lfn
        # this should NOT be needed... but RMS complains!
        bkFile.PFN = lfn
        bkFile.GUID = metaData["GUID"]
        bkFile.Checksum = metaData["Checksum"]
        bkFile.ChecksumType = metaData["ChecksumType"]

        regFile.addFile(bkFile)
        res = self.request.addOperation(regFile)
        if not res["OK"]:
            raise RuntimeError(res["Message"])

    #############################################################################

    # properties
    def set_jobID(self, value):
        if isinstance(value, str):
            if value:
                value = int(value)
            else:
                value = 0
        self._jobID = value

    def get_jobID(self):
        return self._jobID

    jobID = property(get_jobID, set_jobID)

    def _getCurrentOwner(self):
        """Simple function to return current DIRAC username."""
        if "OwnerName" in self.workflow_commons:
            return self.workflow_commons["OwnerName"]

        result = getProxyInfo()

        if not result["OK"]:
            if not self._enableModule():
                return "testUser"
            raise RuntimeError("Could not obtain proxy information")

        if "username" not in result["Value"]:
            raise RuntimeError("Could not get username from proxy")

        return result["Value"]["username"]
